import csv
import os.path

from c_analyzer.parser.declarations import extract_storage
from c_analyzer.variables import known as _common
from c_analyzer.variables.info import Variable

from . import DATA_DIR


# XXX need tests:
# * from_file()
# * look_up_variable()


DATA_FILE = os.path.join(DATA_DIR, 'known.tsv')


def _get_storage(decl, infunc):
    # statics
    if decl.startswith(('Py_LOCAL(', 'Py_LOCAL_INLINE(')):
        return 'static'
    if decl.startswith(('_Py_IDENTIFIER(', '_Py_static_string(')):
        return 'static'
    if decl.startswith('PyDoc_VAR('):
        return 'static'
    if decl.startswith(('SLOT1BINFULL(', 'SLOT1BIN(')):
        return 'static'
    if decl.startswith('WRAP_METHOD('):
        return 'static'
    # public extern
    if decl.startswith('PyAPI_DATA('):
        return 'extern'
    # Fall back to the normal handler.
    return extract_storage(decl, infunc=infunc)


def _handle_var(varid, decl):
#    if varid.name == 'id' and decl == UNKNOWN:
#        # None of these are variables.
#        decl = 'int id';
    storage = _get_storage(decl, varid.funcname)
    return Variable(varid, storage, decl)


def from_file(infile=DATA_FILE, *,
              _from_file=_common.from_file,
              _handle_var=_handle_var,
              ):
    """Return the info for known declarations in the given file."""
    return _from_file(infile, handle_var=_handle_var)


def look_up_variable(varid, knownvars, *,
                     _lookup=_common.look_up_variable,
                     ):
    """Return the known variable matching the given ID.

    "knownvars" is a mapping of ID to Variable.

    "match_files" is used to verify if two filenames point to
    the same file.

    If no match is found then None is returned.
    """
    return _lookup(varid, knownvars)



import sqlite3
import datetime
import time

def adapt_datetime(ts):
    return time.mktime(ts.timetuple())

sqlite3.register_adapter(datetime.datetime, adapt_datetime)

con = sqlite3.connect(":memory:")
cur = con.cursor()

now = datetime.datetime.now()
cur.execute("select ?", (now,))
print(cur.fetchone()[0])

con.close()



import textwrap
import unittest
import sys

from ..util import wrapped_arg_combos, StrProxy
from .. import tool_imports_for_tests
with tool_imports_for_tests():
    from c_analyzer.parser.preprocessor import (
        iter_lines,
        # directives
        parse_directive, PreprocessorDirective,
        Constant, Macro, IfDirective, Include, OtherDirective,
        )


class TestCaseBase(unittest.TestCase):

    maxDiff = None

    def reset(self):
        self._calls = []
        self.errors = None

    @property
    def calls(self):
        try:
            return self._calls
        except AttributeError:
            self._calls = []
            return self._calls

    errors = None

    def try_next_exc(self):
        if not self.errors:
            return
        if exc := self.errors.pop(0):
            raise exc

    def check_calls(self, *expected):
        self.assertEqual(self.calls, list(expected))
        self.assertEqual(self.errors or [], [])


class IterLinesTests(TestCaseBase):

    parsed = None

    def check_calls(self, *expected):
        super().check_calls(*expected)
        self.assertEqual(self.parsed or [], [])

    def _parse_directive(self, line):
        self.calls.append(
                ('_parse_directive', line))
        self.try_next_exc()
        return self.parsed.pop(0)

    def test_no_lines(self):
        lines = []

        results = list(
                iter_lines(lines, _parse_directive=self._parse_directive))

        self.assertEqual(results, [])
        self.check_calls()

    def test_no_directives(self):
        lines = textwrap.dedent('''

            // xyz
            typedef enum {
                SPAM
                EGGS
            } kind;

            struct info {
                kind kind;
                int status;
            };

            typedef struct spam {
                struct info info;
            } myspam;

            static int spam = 0;

            /**
             * ...
             */
            static char *
            get_name(int arg,
                     char *default,
                     )
            {
                return default
            }

            int check(void) {
                return 0;
            }

            ''')[1:-1].splitlines()
        expected = [(lno, line, None, ())
                    for lno, line in enumerate(lines, 1)]
        expected[1] = (2, ' ', None, ())
        expected[20] = (21, ' ', None, ())
        del expected[19]
        del expected[18]

        results = list(
                iter_lines(lines, _parse_directive=self._parse_directive))

        self.assertEqual(results, expected)
        self.check_calls()

    def test_single_directives(self):
        tests = [
            ('#include <stdio>', Include('<stdio>')),
            ('#define SPAM 1', Constant('SPAM', '1')),
            ('#define SPAM() 1', Macro('SPAM', (), '1')),
            ('#define SPAM(a, b) a = b;', Macro('SPAM', ('a', 'b'), 'a = b;')),
            ('#if defined(SPAM)', IfDirective('if', 'defined(SPAM)')),
            ('#ifdef SPAM', IfDirective('ifdef', 'SPAM')),
            ('#ifndef SPAM', IfDirective('ifndef', 'SPAM')),
            ('#elseif defined(SPAM)', IfDirective('elseif', 'defined(SPAM)')),
            ('#else', OtherDirective('else', None)),
            ('#endif', OtherDirective('endif', None)),
            ('#error ...', OtherDirective('error', '...')),
            ('#warning ...', OtherDirective('warning', '...')),
            ('#__FILE__ ...', OtherDirective('__FILE__', '...')),
            ('#__LINE__ ...', OtherDirective('__LINE__', '...')),
            ('#__DATE__ ...', OtherDirective('__DATE__', '...')),
            ('#__TIME__ ...', OtherDirective('__TIME__', '...')),
            ('#__TIMESTAMP__ ...', OtherDirective('__TIMESTAMP__', '...')),
            ]
        for line, directive in tests:
            with self.subTest(line):
                self.reset()
                self.parsed = [
                    directive,
                    ]
                text = textwrap.dedent('''
                    static int spam = 0;
                    {}
                    static char buffer[256];
                    ''').strip().format(line)
                lines = text.strip().splitlines()

                results = list(
                        iter_lines(lines, _parse_directive=self._parse_directive))

                self.assertEqual(results, [
                    (1, 'static int spam = 0;', None, ()),
                    (2, line, directive, ()),
                    ((3, 'static char buffer[256];', None, ('defined(SPAM)',))
                     if directive.kind in ('if', 'ifdef', 'elseif')
                     else (3, 'static char buffer[256];', None, ('! defined(SPAM)',))
                     if directive.kind == 'ifndef'
                     else (3, 'static char buffer[256];', None, ())),
                    ])
                self.check_calls(
                        ('_parse_directive', line),
                        )

    def test_directive_whitespace(self):
        line = ' # define  eggs  (  a  ,  b  )  {  a  =  b  ;  }  '
        directive = Macro('eggs', ('a', 'b'), '{ a = b; }')
        self.parsed = [
            directive,
            ]
        lines = [line]

        results = list(
                iter_lines(lines, _parse_directive=self._parse_directive))

        self.assertEqual(results, [
            (1, line, directive, ()),
            ])
        self.check_calls(
                ('_parse_directive', '#define eggs ( a , b ) { a = b ; }'),
                )

    @unittest.skipIf(sys.platform == 'win32', 'needs fix under Windows')
    def test_split_lines(self):
        directive = Macro('eggs', ('a', 'b'), '{ a = b; }')
        self.parsed = [
            directive,
            ]
        text = textwrap.dedent(r'''
            static int spam = 0;
            #define eggs(a, b) \
                { \
                    a = b; \
                }
            static char buffer[256];
            ''').strip()
        lines = [line + '\n' for line in text.splitlines()]
        lines[-1] = lines[-1][:-1]

        results = list(
                iter_lines(lines, _parse_directive=self._parse_directive))

        self.assertEqual(results, [
            (1, 'static int spam = 0;\n', None, ()),
            (5, '#define eggs(a, b)      {          a = b;      }\n', directive, ()),
            (6, 'static char buffer[256];', None, ()),
            ])
        self.check_calls(
                ('_parse_directive', '#define eggs(a, b) { a = b; }'),
                )

    def test_nested_conditions(self):
        directives = [
            IfDirective('ifdef', 'SPAM'),
            IfDirective('if', 'SPAM == 1'),
            IfDirective('elseif', 'SPAM == 2'),
            OtherDirective('else', None),
            OtherDirective('endif', None),
            OtherDirective('endif', None),
            ]
        self.parsed = list(directives)
        text = textwrap.dedent(r'''
            static int spam = 0;

            #ifdef SPAM
            static int start = 0;
            #  if SPAM == 1
            static char buffer[10];
            #  elif SPAM == 2
            static char buffer[100];
            #  else
            static char buffer[256];
            #  endif
            static int end = 0;
            #endif

            static int eggs = 0;
            ''').strip()
        lines = [line for line in text.splitlines() if line.strip()]

        results = list(
                iter_lines(lines, _parse_directive=self._parse_directive))

        self.assertEqual(results, [
            (1, 'static int spam = 0;', None, ()),
            (2, '#ifdef SPAM', directives[0], ()),
            (3, 'static int start = 0;', None, ('defined(SPAM)',)),
            (4, '#  if SPAM == 1', directives[1], ('defined(SPAM)',)),
            (5, 'static char buffer[10];', None, ('defined(SPAM)', 'SPAM == 1')),
            (6, '#  elif SPAM == 2', directives[2], ('defined(SPAM)', 'SPAM == 1')),
            (7, 'static char buffer[100];', None, ('defined(SPAM)', '! (SPAM == 1)', 'SPAM == 2')),
            (8, '#  else', directives[3], ('defined(SPAM)', '! (SPAM == 1)', 'SPAM == 2')),
            (9, 'static char buffer[256];', None, ('defined(SPAM)', '! (SPAM == 1)', '! (SPAM == 2)')),
            (10, '#  endif', directives[4], ('defined(SPAM)', '! (SPAM == 1)', '! (SPAM == 2)')),
            (11, 'static int end = 0;', None, ('defined(SPAM)',)),
            (12, '#endif', directives[5], ('defined(SPAM)',)),
            (13, 'static int eggs = 0;', None, ()),
            ])
        self.check_calls(
                ('_parse_directive', '#ifdef SPAM'),
                ('_parse_directive', '#if SPAM == 1'),
                ('_parse_directive', '#elif SPAM == 2'),
                ('_parse_directive', '#else'),
                ('_parse_directive', '#endif'),
                ('_parse_directive', '#endif'),
                )

    def test_split_blocks(self):
        directives = [
            IfDirective('ifdef', 'SPAM'),
            OtherDirective('else', None),
            OtherDirective('endif', None),
            ]
        self.parsed = list(directives)
        text = textwrap.dedent(r'''
            void str_copy(char *buffer, *orig);

            int init(char *name) {
                static int initialized = 0;
                if (initialized) {
                    return 0;
                }
            #ifdef SPAM
                static char buffer[10];
                str_copy(buffer, char);
            }

            void copy(char *buffer, *orig) {
                strncpy(buffer, orig, 9);
                buffer[9] = 0;
            }

            #else
                static char buffer[256];
                str_copy(buffer, char);
            }

            void copy(char *buffer, *orig) {
                strcpy(buffer, orig);
            }

            #endif
            ''').strip()
        lines = [line for line in text.splitlines() if line.strip()]

        results = list(
                iter_lines(lines, _parse_directive=self._parse_directive))

        self.assertEqual(results, [
            (1, 'void str_copy(char *buffer, *orig);', None, ()),
            (2, 'int init(char *name) {', None, ()),
            (3, '    static int initialized = 0;', None, ()),
            (4, '    if (initialized) {', None, ()),
            (5, '        return 0;', None, ()),
            (6, '    }', None, ()),

            (7, '#ifdef SPAM', directives[0], ()),

            (8, '    static char buffer[10];', None, ('defined(SPAM)',)),
            (9, '    str_copy(buffer, char);', None, ('defined(SPAM)',)),
            (10, '}', None, ('defined(SPAM)',)),
            (11, 'void copy(char *buffer, *orig) {', None, ('defined(SPAM)',)),
            (12, '    strncpy(buffer, orig, 9);', None, ('defined(SPAM)',)),
            (13, '    buffer[9] = 0;', None, ('defined(SPAM)',)),
            (14, '}', None, ('defined(SPAM)',)),

            (15, '#else', directives[1], ('defined(SPAM)',)),

            (16, '    static char buffer[256];', None, ('! (defined(SPAM))',)),
            (17, '    str_copy(buffer, char);', None, ('! (defined(SPAM))',)),
            (18, '}', None, ('! (defined(SPAM))',)),
            (19, 'void copy(char *buffer, *orig) {', None, ('! (defined(SPAM))',)),
            (20, '    strcpy(buffer, orig);', None, ('! (defined(SPAM))',)),
            (21, '}', None, ('! (defined(SPAM))',)),

            (22, '#endif', directives[2], ('! (defined(SPAM))',)),
            ])
        self.check_calls(
                ('_parse_directive', '#ifdef SPAM'),
                ('_parse_directive', '#else'),
                ('_parse_directive', '#endif'),
                )

    @unittest.skipIf(sys.platform == 'win32', 'needs fix under Windows')
    def test_basic(self):
        directives = [
            Include('<stdio.h>'),
            IfDirective('ifdef', 'SPAM'),
            IfDirective('if', '! defined(HAM) || !HAM'),
            Constant('HAM', '0'),
            IfDirective('elseif', 'HAM < 0'),
            Constant('HAM', '-1'),
            OtherDirective('else', None),
            OtherDirective('endif', None),
            OtherDirective('endif', None),
            IfDirective('if', 'defined(HAM) && (HAM < 0 || ! HAM)'),
            OtherDirective('undef', 'HAM'),
            OtherDirective('endif', None),
            IfDirective('ifndef', 'HAM'),
            OtherDirective('endif', None),
            ]
        self.parsed = list(directives)
        text = textwrap.dedent(r'''
            #include <stdio.h>
            print("begin");
            #ifdef SPAM
               print("spam");
               #if ! defined(HAM) || !HAM
            #      DEFINE HAM 0
               #elseif HAM < 0
            #      DEFINE HAM -1
               #else
                   print("ham HAM");
               #endif
            #endif

            #if defined(HAM) && \
                (HAM < 0 || ! HAM)
              print("ham?");
              #undef HAM
            # endif

            #ifndef HAM
               print("no ham");
            #endif
            print("end");
            ''')[1:-1]
        lines = [line + '\n' for line in text.splitlines()]
        lines[-1] = lines[-1][:-1]

        results = list(
                iter_lines(lines, _parse_directive=self._parse_directive))

        self.assertEqual(results, [
            (1, '#include <stdio.h>\n', Include('<stdio.h>'), ()),
            (2, 'print("begin");\n', None, ()),
            #
            (3, '#ifdef SPAM\n',
                IfDirective('ifdef', 'SPAM'),
                ()),
            (4, '   print("spam");\n',
                None,
                ('defined(SPAM)',)),
            (5, '   #if ! defined(HAM) || !HAM\n',
                IfDirective('if', '! defined(HAM) || !HAM'),
                ('defined(SPAM)',)),
            (6, '#      DEFINE HAM 0\n',
                Constant('HAM', '0'),
                ('defined(SPAM)', '! defined(HAM) || !HAM')),
            (7, '   #elseif HAM < 0\n',
                IfDirective('elseif', 'HAM < 0'),
                ('defined(SPAM)', '! defined(HAM) || !HAM')),
            (8, '#      DEFINE HAM -1\n',
                Constant('HAM', '-1'),
                ('defined(SPAM)', '! (! defined(HAM) || !HAM)', 'HAM < 0')),
            (9, '   #else\n',
                OtherDirective('else', None),
                ('defined(SPAM)', '! (! defined(HAM) || !HAM)', 'HAM < 0')),
            (10, '       print("ham HAM");\n',
                None,
                ('defined(SPAM)', '! (! defined(HAM) || !HAM)', '! (HAM < 0)')),
            (11, '   #endif\n',
                OtherDirective('endif', None),
                ('defined(SPAM)', '! (! defined(HAM) || !HAM)', '! (HAM < 0)')),
            (12, '#endif\n',
                OtherDirective('endif', None),
                ('defined(SPAM)',)),
            #
            (13, '\n', None, ()),
            #
            (15, '#if defined(HAM) &&      (HAM < 0 || ! HAM)\n',
                IfDirective('if', 'defined(HAM) && (HAM < 0 || ! HAM)'),
                ()),
            (16, '  print("ham?");\n',
                None,
                ('defined(HAM) && (HAM < 0 || ! HAM)',)),
            (17, '  #undef HAM\n',
                OtherDirective('undef', 'HAM'),
                ('defined(HAM) && (HAM < 0 || ! HAM)',)),
            (18, '# endif\n',
                OtherDirective('endif', None),
                ('defined(HAM) && (HAM < 0 || ! HAM)',)),
            #
            (19, '\n', None, ()),
            #
            (20, '#ifndef HAM\n',
                IfDirective('ifndef', 'HAM'),
                ()),
            (21, '   print("no ham");\n',
                None,
                ('! defined(HAM)',)),
            (22, '#endif\n',
                OtherDirective('endif', None),
                ('! defined(HAM)',)),
            #
            (23, 'print("end");', None, ()),
            ])

    @unittest.skipIf(sys.platform == 'win32', 'needs fix under Windows')
    def test_typical(self):
        # We use Include/compile.h from commit 66c4f3f38b86.  It has
        # a good enough mix of code without being too large.
        directives = [
            IfDirective('ifndef', 'Py_COMPILE_H'),
            Constant('Py_COMPILE_H', None),

            IfDirective('ifndef', 'Py_LIMITED_API'),

            Include('"code.h"'),

            IfDirective('ifdef', '__cplusplus'),
            OtherDirective('endif', None),

            Constant('PyCF_MASK', '(CO_FUTURE_DIVISION | CO_FUTURE_ABSOLUTE_IMPORT | CO_FUTURE_WITH_STATEMENT | CO_FUTURE_PRINT_FUNCTION | CO_FUTURE_UNICODE_LITERALS | CO_FUTURE_BARRY_AS_BDFL | CO_FUTURE_GENERATOR_STOP | CO_FUTURE_ANNOTATIONS)'),
            Constant('PyCF_MASK_OBSOLETE', '(CO_NESTED)'),
            Constant('PyCF_SOURCE_IS_UTF8', ' 0x0100'),
            Constant('PyCF_DONT_IMPLY_DEDENT', '0x0200'),
            Constant('PyCF_ONLY_AST', '0x0400'),
            Constant('PyCF_IGNORE_COOKIE', '0x0800'),
            Constant('PyCF_TYPE_COMMENTS', '0x1000'),
            Constant('PyCF_ALLOW_TOP_LEVEL_AWAIT', '0x2000'),

            IfDirective('ifndef', 'Py_LIMITED_API'),
            OtherDirective('endif', None),

            Constant('FUTURE_NESTED_SCOPES', '"nested_scopes"'),
            Constant('FUTURE_GENERATORS', '"generators"'),
            Constant('FUTURE_DIVISION', '"division"'),
            Constant('FUTURE_ABSOLUTE_IMPORT', '"absolute_import"'),
            Constant('FUTURE_WITH_STATEMENT', '"with_statement"'),
            Constant('FUTURE_PRINT_FUNCTION', '"print_function"'),
            Constant('FUTURE_UNICODE_LITERALS', '"unicode_literals"'),
            Constant('FUTURE_BARRY_AS_BDFL', '"barry_as_FLUFL"'),
            Constant('FUTURE_GENERATOR_STOP', '"generator_stop"'),
            Constant('FUTURE_ANNOTATIONS', '"annotations"'),

            Macro('PyAST_Compile', ('mod', 's', 'f', 'ar'), 'PyAST_CompileEx(mod, s, f, -1, ar)'),

            Constant('PY_INVALID_STACK_EFFECT', 'INT_MAX'),

            IfDirective('ifdef', '__cplusplus'),
            OtherDirective('endif', None),

            OtherDirective('endif', None),  # ifndef Py_LIMITED_API

            Constant('Py_single_input', '256'),
            Constant('Py_file_input', '257'),
            Constant('Py_eval_input', '258'),
            Constant('Py_func_type_input', '345'),

            OtherDirective('endif', None),  # ifndef Py_COMPILE_H
            ]
        self.parsed = list(directives)
        text = textwrap.dedent(r'''
            #ifndef Py_COMPILE_H
            #define Py_COMPILE_H

            #ifndef Py_LIMITED_API
            #include "code.h"

            #ifdef __cplusplus
            extern "C" {
            #endif

            /* Public interface */
            struct _node; /* Declare the existence of this type */
            PyAPI_FUNC(PyCodeObject *) PyNode_Compile(struct _node *, const char *);
            /* XXX (ncoghlan): Unprefixed type name in a public API! */

            #define PyCF_MASK (CO_FUTURE_DIVISION | CO_FUTURE_ABSOLUTE_IMPORT | \
                               CO_FUTURE_WITH_STATEMENT | CO_FUTURE_PRINT_FUNCTION | \
                               CO_FUTURE_UNICODE_LITERALS | CO_FUTURE_BARRY_AS_BDFL | \
                               CO_FUTURE_GENERATOR_STOP | CO_FUTURE_ANNOTATIONS)
            #define PyCF_MASK_OBSOLETE (CO_NESTED)
            #define PyCF_SOURCE_IS_UTF8  0x0100
            #define PyCF_DONT_IMPLY_DEDENT 0x0200
            #define PyCF_ONLY_AST 0x0400
            #define PyCF_IGNORE_COOKIE 0x0800
            #define PyCF_TYPE_COMMENTS 0x1000
            #define PyCF_ALLOW_TOP_LEVEL_AWAIT 0x2000

            #ifndef Py_LIMITED_API
            typedef struct {
                int cf_flags;  /* bitmask of CO_xxx flags relevant to future */
                int cf_feature_version;  /* minor Python version (PyCF_ONLY_AST) */
            } PyCompilerFlags;
            #endif

            /* Future feature support */

            typedef struct {
                int ff_features;      /* flags set by future statements */
                int ff_lineno;        /* line number of last future statement */
            } PyFutureFeatures;

            #define FUTURE_NESTED_SCOPES "nested_scopes"
            #define FUTURE_GENERATORS "generators"
            #define FUTURE_DIVISION "division"
            #define FUTURE_ABSOLUTE_IMPORT "absolute_import"
            #define FUTURE_WITH_STATEMENT "with_statement"
            #define FUTURE_PRINT_FUNCTION "print_function"
            #define FUTURE_UNICODE_LITERALS "unicode_literals"
            #define FUTURE_BARRY_AS_BDFL "barry_as_FLUFL"
            #define FUTURE_GENERATOR_STOP "generator_stop"
            #define FUTURE_ANNOTATIONS "annotations"

            struct _mod; /* Declare the existence of this type */
            #define PyAST_Compile(mod, s, f, ar) PyAST_CompileEx(mod, s, f, -1, ar)
            PyAPI_FUNC(PyCodeObject *) PyAST_CompileEx(
                struct _mod *mod,
                const char *filename,       /* decoded from the filesystem encoding */
                PyCompilerFlags *flags,
                int optimize,
                PyArena *arena);
            PyAPI_FUNC(PyCodeObject *) PyAST_CompileObject(
                struct _mod *mod,
                PyObject *filename,
                PyCompilerFlags *flags,
                int optimize,
                PyArena *arena);
            PyAPI_FUNC(PyFutureFeatures *) PyFuture_FromAST(
                struct _mod * mod,
                const char *filename        /* decoded from the filesystem encoding */
                );
            PyAPI_FUNC(PyFutureFeatures *) PyFuture_FromASTObject(
                struct _mod * mod,
                PyObject *filename
                );

            /* _Py_Mangle is defined in compile.c */
            PyAPI_FUNC(PyObject*) _Py_Mangle(PyObject *p, PyObject *name);

            #define PY_INVALID_STACK_EFFECT INT_MAX
            PyAPI_FUNC(int) PyCompile_OpcodeStackEffect(int opcode, int oparg);
            PyAPI_FUNC(int) PyCompile_OpcodeStackEffectWithJump(int opcode, int oparg, int jump);

            PyAPI_FUNC(int) _PyAST_Optimize(struct _mod *, PyArena *arena, int optimize);

            #ifdef __cplusplus
            }
            #endif

            #endif /* !Py_LIMITED_API */

            /* These definitions must match corresponding definitions in graminit.h. */
            #define Py_single_input 256
            #define Py_file_input 257
            #define Py_eval_input 258
            #define Py_func_type_input 345

            #endif /* !Py_COMPILE_H */
            ''').strip()
        lines = [line + '\n' for line in text.splitlines()]
        lines[-1] = lines[-1][:-1]

        results = list(
                iter_lines(lines, _parse_directive=self._parse_directive))

        self.assertEqual(results, [
            (1, '#ifndef Py_COMPILE_H\n',
                IfDirective('ifndef', 'Py_COMPILE_H'),
                ()),
            (2, '#define Py_COMPILE_H\n',
                Constant('Py_COMPILE_H', None),
                ('! defined(Py_COMPILE_H)',)),
            (3, '\n',
                None,
                ('! defined(Py_COMPILE_H)',)),
            (4, '#ifndef Py_LIMITED_API\n',
                IfDirective('ifndef', 'Py_LIMITED_API'),
                ('! defined(Py_COMPILE_H)',)),
            (5, '#include "code.h"\n',
                Include('"code.h"'),
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (6, '\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (7, '#ifdef __cplusplus\n',
                IfDirective('ifdef', '__cplusplus'),
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (8, 'extern "C" {\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)', 'defined(__cplusplus)')),
            (9, '#endif\n',
                OtherDirective('endif', None),
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)', 'defined(__cplusplus)')),
            (10, '\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (11, ' \n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (12, 'struct _node;  \n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (13, 'PyAPI_FUNC(PyCodeObject *) PyNode_Compile(struct _node *, const char *);\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (14, ' \n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (15, '\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (19, '#define PyCF_MASK (CO_FUTURE_DIVISION | CO_FUTURE_ABSOLUTE_IMPORT |                     CO_FUTURE_WITH_STATEMENT | CO_FUTURE_PRINT_FUNCTION |                     CO_FUTURE_UNICODE_LITERALS | CO_FUTURE_BARRY_AS_BDFL |                     CO_FUTURE_GENERATOR_STOP | CO_FUTURE_ANNOTATIONS)\n',
                Constant('PyCF_MASK', '(CO_FUTURE_DIVISION | CO_FUTURE_ABSOLUTE_IMPORT | CO_FUTURE_WITH_STATEMENT | CO_FUTURE_PRINT_FUNCTION | CO_FUTURE_UNICODE_LITERALS | CO_FUTURE_BARRY_AS_BDFL | CO_FUTURE_GENERATOR_STOP | CO_FUTURE_ANNOTATIONS)'),
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (20, '#define PyCF_MASK_OBSOLETE (CO_NESTED)\n',
                Constant('PyCF_MASK_OBSOLETE', '(CO_NESTED)'),
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (21, '#define PyCF_SOURCE_IS_UTF8  0x0100\n',
                Constant('PyCF_SOURCE_IS_UTF8', ' 0x0100'),
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (22, '#define PyCF_DONT_IMPLY_DEDENT 0x0200\n',
                Constant('PyCF_DONT_IMPLY_DEDENT', '0x0200'),
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (23, '#define PyCF_ONLY_AST 0x0400\n',
                Constant('PyCF_ONLY_AST', '0x0400'),
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (24, '#define PyCF_IGNORE_COOKIE 0x0800\n',
                Constant('PyCF_IGNORE_COOKIE', '0x0800'),
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (25, '#define PyCF_TYPE_COMMENTS 0x1000\n',
                Constant('PyCF_TYPE_COMMENTS', '0x1000'),
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (26, '#define PyCF_ALLOW_TOP_LEVEL_AWAIT 0x2000\n',
                Constant('PyCF_ALLOW_TOP_LEVEL_AWAIT', '0x2000'),
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (27, '\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (28, '#ifndef Py_LIMITED_API\n',
                IfDirective('ifndef', 'Py_LIMITED_API'),
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (29, 'typedef struct {\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)', '! defined(Py_LIMITED_API)')),
            (30, '    int cf_flags;   \n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)', '! defined(Py_LIMITED_API)')),
            (31, '    int cf_feature_version;   \n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)', '! defined(Py_LIMITED_API)')),
            (32, '} PyCompilerFlags;\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)', '! defined(Py_LIMITED_API)')),
            (33, '#endif\n',
                OtherDirective('endif', None),
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)', '! defined(Py_LIMITED_API)')),
            (34, '\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (35, ' \n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (36, '\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (37, 'typedef struct {\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (38, '    int ff_features;       \n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (39, '    int ff_lineno;         \n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (40, '} PyFutureFeatures;\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (41, '\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (42, '#define FUTURE_NESTED_SCOPES "nested_scopes"\n',
                Constant('FUTURE_NESTED_SCOPES', '"nested_scopes"'),
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (43, '#define FUTURE_GENERATORS "generators"\n',
                Constant('FUTURE_GENERATORS', '"generators"'),
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (44, '#define FUTURE_DIVISION "division"\n',
                Constant('FUTURE_DIVISION', '"division"'),
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (45, '#define FUTURE_ABSOLUTE_IMPORT "absolute_import"\n',
                Constant('FUTURE_ABSOLUTE_IMPORT', '"absolute_import"'),
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (46, '#define FUTURE_WITH_STATEMENT "with_statement"\n',
                Constant('FUTURE_WITH_STATEMENT', '"with_statement"'),
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (47, '#define FUTURE_PRINT_FUNCTION "print_function"\n',
                Constant('FUTURE_PRINT_FUNCTION', '"print_function"'),
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (48, '#define FUTURE_UNICODE_LITERALS "unicode_literals"\n',
                Constant('FUTURE_UNICODE_LITERALS', '"unicode_literals"'),
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (49, '#define FUTURE_BARRY_AS_BDFL "barry_as_FLUFL"\n',
                Constant('FUTURE_BARRY_AS_BDFL', '"barry_as_FLUFL"'),
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (50, '#define FUTURE_GENERATOR_STOP "generator_stop"\n',
                Constant('FUTURE_GENERATOR_STOP', '"generator_stop"'),
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (51, '#define FUTURE_ANNOTATIONS "annotations"\n',
                Constant('FUTURE_ANNOTATIONS', '"annotations"'),
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (52, '\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (53, 'struct _mod;  \n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (54, '#define PyAST_Compile(mod, s, f, ar) PyAST_CompileEx(mod, s, f, -1, ar)\n',
                Macro('PyAST_Compile', ('mod', 's', 'f', 'ar'), 'PyAST_CompileEx(mod, s, f, -1, ar)'),
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (55, 'PyAPI_FUNC(PyCodeObject *) PyAST_CompileEx(\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (56, '    struct _mod *mod,\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (57, '    const char *filename,        \n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (58, '    PyCompilerFlags *flags,\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (59, '    int optimize,\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (60, '    PyArena *arena);\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (61, 'PyAPI_FUNC(PyCodeObject *) PyAST_CompileObject(\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (62, '    struct _mod *mod,\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (63, '    PyObject *filename,\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (64, '    PyCompilerFlags *flags,\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (65, '    int optimize,\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (66, '    PyArena *arena);\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (67, 'PyAPI_FUNC(PyFutureFeatures *) PyFuture_FromAST(\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (68, '    struct _mod * mod,\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (69, '    const char *filename         \n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (70, '    );\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (71, 'PyAPI_FUNC(PyFutureFeatures *) PyFuture_FromASTObject(\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (72, '    struct _mod * mod,\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (73, '    PyObject *filename\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (74, '    );\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (75, '\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (76, ' \n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (77, 'PyAPI_FUNC(PyObject*) _Py_Mangle(PyObject *p, PyObject *name);\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (78, '\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (79, '#define PY_INVALID_STACK_EFFECT INT_MAX\n',
                Constant('PY_INVALID_STACK_EFFECT', 'INT_MAX'),
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (80, 'PyAPI_FUNC(int) PyCompile_OpcodeStackEffect(int opcode, int oparg);\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (81, 'PyAPI_FUNC(int) PyCompile_OpcodeStackEffectWithJump(int opcode, int oparg, int jump);\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (82, '\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (83, 'PyAPI_FUNC(int) _PyAST_Optimize(struct _mod *, PyArena *arena, int optimize);\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (84, '\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (85, '#ifdef __cplusplus\n',
                IfDirective('ifdef', '__cplusplus'),
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (86, '}\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)', 'defined(__cplusplus)')),
            (87, '#endif\n',
                OtherDirective('endif', None),
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)', 'defined(__cplusplus)')),
            (88, '\n',
                None,
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (89, '#endif  \n',
                OtherDirective('endif', None),
                ('! defined(Py_COMPILE_H)', '! defined(Py_LIMITED_API)')),
            (90, '\n',
                None,
                ('! defined(Py_COMPILE_H)',)),
            (91, ' \n',
                None,
                ('! defined(Py_COMPILE_H)',)),
            (92, '#define Py_single_input 256\n',
                Constant('Py_single_input', '256'),
                ('! defined(Py_COMPILE_H)',)),
            (93, '#define Py_file_input 257\n',
                Constant('Py_file_input', '257'),
                ('! defined(Py_COMPILE_H)',)),
            (94, '#define Py_eval_input 258\n',
                Constant('Py_eval_input', '258'),
                ('! defined(Py_COMPILE_H)',)),
            (95, '#define Py_func_type_input 345\n',
                Constant('Py_func_type_input', '345'),
                ('! defined(Py_COMPILE_H)',)),
            (96, '\n',
                None,
                ('! defined(Py_COMPILE_H)',)),
            (97, '#endif  ',
                OtherDirective('endif', None),
                ('! defined(Py_COMPILE_H)',)),
            ])
        self.check_calls(
                ('_parse_directive', '#ifndef Py_COMPILE_H'),
                ('_parse_directive', '#define Py_COMPILE_H'),
                ('_parse_directive', '#ifndef Py_LIMITED_API'),
                ('_parse_directive', '#include "code.h"'),
                ('_parse_directive', '#ifdef __cplusplus'),
                ('_parse_directive', '#endif'),
                ('_parse_directive', '#define PyCF_MASK (CO_FUTURE_DIVISION | CO_FUTURE_ABSOLUTE_IMPORT | CO_FUTURE_WITH_STATEMENT | CO_FUTURE_PRINT_FUNCTION | CO_FUTURE_UNICODE_LITERALS | CO_FUTURE_BARRY_AS_BDFL | CO_FUTURE_GENERATOR_STOP | CO_FUTURE_ANNOTATIONS)'),
                ('_parse_directive', '#define PyCF_MASK_OBSOLETE (CO_NESTED)'),
                ('_parse_directive', '#define PyCF_SOURCE_IS_UTF8 0x0100'),
                ('_parse_directive', '#define PyCF_DONT_IMPLY_DEDENT 0x0200'),
                ('_parse_directive', '#define PyCF_ONLY_AST 0x0400'),
                ('_parse_directive', '#define PyCF_IGNORE_COOKIE 0x0800'),
                ('_parse_directive', '#define PyCF_TYPE_COMMENTS 0x1000'),
                ('_parse_directive', '#define PyCF_ALLOW_TOP_LEVEL_AWAIT 0x2000'),
                ('_parse_directive', '#ifndef Py_LIMITED_API'),
                ('_parse_directive', '#endif'),
                ('_parse_directive', '#define FUTURE_NESTED_SCOPES "nested_scopes"'),
                ('_parse_directive', '#define FUTURE_GENERATORS "generators"'),
                ('_parse_directive', '#define FUTURE_DIVISION "division"'),
                ('_parse_directive', '#define FUTURE_ABSOLUTE_IMPORT "absolute_import"'),
                ('_parse_directive', '#define FUTURE_WITH_STATEMENT "with_statement"'),
                ('_parse_directive', '#define FUTURE_PRINT_FUNCTION "print_function"'),
                ('_parse_directive', '#define FUTURE_UNICODE_LITERALS "unicode_literals"'),
                ('_parse_directive', '#define FUTURE_BARRY_AS_BDFL "barry_as_FLUFL"'),
                ('_parse_directive', '#define FUTURE_GENERATOR_STOP "generator_stop"'),
                ('_parse_directive', '#define FUTURE_ANNOTATIONS "annotations"'),
                ('_parse_directive', '#define PyAST_Compile(mod, s, f, ar) PyAST_CompileEx(mod, s, f, -1, ar)'),
                ('_parse_directive', '#define PY_INVALID_STACK_EFFECT INT_MAX'),
                ('_parse_directive', '#ifdef __cplusplus'),
                ('_parse_directive', '#endif'),
                ('_parse_directive', '#endif'),
                ('_parse_directive', '#define Py_single_input 256'),
                ('_parse_directive', '#define Py_file_input 257'),
                ('_parse_directive', '#define Py_eval_input 258'),
                ('_parse_directive', '#define Py_func_type_input 345'),
                ('_parse_directive', '#endif'),
                )


class ParseDirectiveTests(unittest.TestCase):

    def test_directives(self):
        tests = [
            # includes
            ('#include "internal/pycore_pystate.h"', Include('"internal/pycore_pystate.h"')),
            ('#include <stdio>', Include('<stdio>')),

            # defines
            ('#define SPAM int', Constant('SPAM', 'int')),
            ('#define SPAM', Constant('SPAM', '')),
            ('#define SPAM(x, y) run(x, y)', Macro('SPAM', ('x', 'y'), 'run(x, y)')),
            ('#undef SPAM', None),

            # conditionals
            ('#if SPAM', IfDirective('if', 'SPAM')),
            # XXX complex conditionls
            ('#ifdef SPAM', IfDirective('ifdef', 'SPAM')),
            ('#ifndef SPAM', IfDirective('ifndef', 'SPAM')),
            ('#elseif SPAM', IfDirective('elseif', 'SPAM')),
            # XXX complex conditionls
            ('#else', OtherDirective('else', '')),
            ('#endif', OtherDirective('endif', '')),

            # other
            ('#error oops!', None),
            ('#warning oops!', None),
            ('#pragma ...', None),
            ('#__FILE__ ...', None),
            ('#__LINE__ ...', None),
            ('#__DATE__ ...', None),
            ('#__TIME__ ...', None),
            ('#__TIMESTAMP__ ...', None),

            # extra whitespace
            (' # include  <stdio> ', Include('<stdio>')),
            ('#else  ', OtherDirective('else', '')),
            ('#endif  ', OtherDirective('endif', '')),
            ('#define SPAM int  ', Constant('SPAM', 'int')),
            ('#define SPAM  ', Constant('SPAM', '')),
            ]
        for line, expected in tests:
            if expected is None:
                kind, _, text = line[1:].partition(' ')
                expected = OtherDirective(kind, text)
            with self.subTest(line):
                directive = parse_directive(line)

                self.assertEqual(directive, expected)

    def test_bad_directives(self):
        tests = [
            # valid directives with bad text
            '#define 123',
            '#else spam',
            '#endif spam',
            ]
        for kind in PreprocessorDirective.KINDS:
            # missing leading "#"
            tests.append(kind)
            if kind in ('else', 'endif'):
                continue
            # valid directives with missing text
            tests.append('#' + kind)
            tests.append('#' + kind + ' ')
        for line in tests:
            with self.subTest(line):
                with self.assertRaises(ValueError):
                    parse_directive(line)

    def test_not_directives(self):
        tests = [
            '',
            ' ',
            'directive',
            'directive?',
            '???',
            ]
        for line in tests:
            with self.subTest(line):
                with self.assertRaises(ValueError):
                    parse_directive(line)


class ConstantTests(unittest.TestCase):

    def test_type(self):
        directive = Constant('SPAM', '123')

        self.assertIs(type(directive), Constant)
        self.assertIsInstance(directive, PreprocessorDirective)

    def test_attrs(self):
        d = Constant('SPAM', '123')
        kind, name, value = d.kind, d.name, d.value

        self.assertEqual(kind, 'define')
        self.assertEqual(name, 'SPAM')
        self.assertEqual(value, '123')

    def test_text(self):
        tests = [
            (('SPAM', '123'), 'SPAM 123'),
            (('SPAM',), 'SPAM'),
            ]
        for args, expected in tests:
            with self.subTest(args):
                d = Constant(*args)
                text = d.text

                self.assertEqual(text, expected)

    def test_iter(self):
        kind, name, value = Constant('SPAM', '123')

        self.assertEqual(kind, 'define')
        self.assertEqual(name, 'SPAM')
        self.assertEqual(value, '123')

    def test_defaults(self):
        kind, name, value = Constant('SPAM')

        self.assertEqual(kind, 'define')
        self.assertEqual(name, 'SPAM')
        self.assertIs(value, None)

    def test_coerce(self):
        tests = []
        # coerced name, value
        for args in wrapped_arg_combos('SPAM', '123'):
            tests.append((args, ('SPAM', '123')))
        # missing name, value
        for name in ('', ' ', None, StrProxy(' '), ()):
            for value in ('', ' ', None, StrProxy(' '), ()):
                tests.append(
                        ((name, value), (None, None)))
        # whitespace
        tests.extend([
            ((' SPAM ', ' 123 '), ('SPAM', '123')),
            ])

        for args, expected in tests:
            with self.subTest(args):
                d = Constant(*args)

                self.assertEqual(d[1:], expected)
                for i, exp in enumerate(expected, start=1):
                    if exp is not None:
                        self.assertIs(type(d[i]), str)

    def test_valid(self):
        tests = [
            ('SPAM', '123'),
            # unusual name
            ('_SPAM_', '123'),
            ('X_1', '123'),
            # unusual value
            ('SPAM', None),
            ]
        for args in tests:
            with self.subTest(args):
                directive = Constant(*args)

                directive.validate()

    def test_invalid(self):
        tests = [
            # invalid name
            ((None, '123'), TypeError),
            (('_', '123'), ValueError),
            (('1', '123'), ValueError),
            (('_1_', '123'), ValueError),
            # There is no invalid value (including None).
            ]
        for args, exctype in tests:
            with self.subTest(args):
                directive = Constant(*args)

                with self.assertRaises(exctype):
                    directive.validate()


class MacroTests(unittest.TestCase):

    def test_type(self):
        directive = Macro('SPAM', ('x', 'y'), '123')

        self.assertIs(type(directive), Macro)
        self.assertIsInstance(directive, PreprocessorDirective)

    def test_attrs(self):
        d = Macro('SPAM', ('x', 'y'), '123')
        kind, name, args, body = d.kind, d.name, d.args, d.body

        self.assertEqual(kind, 'define')
        self.assertEqual(name, 'SPAM')
        self.assertEqual(args, ('x', 'y'))
        self.assertEqual(body, '123')

    def test_text(self):
        tests = [
            (('SPAM', ('x', 'y'), '123'), 'SPAM(x, y) 123'),
            (('SPAM', ('x', 'y'),), 'SPAM(x, y)'),
            ]
        for args, expected in tests:
            with self.subTest(args):
                d = Macro(*args)
                text = d.text

                self.assertEqual(text, expected)

    def test_iter(self):
        kind, name, args, body = Macro('SPAM', ('x', 'y'), '123')

        self.assertEqual(kind, 'define')
        self.assertEqual(name, 'SPAM')
        self.assertEqual(args, ('x', 'y'))
        self.assertEqual(body, '123')

    def test_defaults(self):
        kind, name, args, body = Macro('SPAM', ('x', 'y'))

        self.assertEqual(kind, 'define')
        self.assertEqual(name, 'SPAM')
        self.assertEqual(args, ('x', 'y'))
        self.assertIs(body, None)

    def test_coerce(self):
        tests = []
        # coerce name and body
        for args in wrapped_arg_combos('SPAM', ('x', 'y'), '123'):
            tests.append(
                    (args, ('SPAM', ('x', 'y'), '123')))
        # coerce args
        tests.extend([
            (('SPAM', 'x', '123'),
             ('SPAM', ('x',), '123')),
            (('SPAM', 'x,y', '123'),
             ('SPAM', ('x', 'y'), '123')),
            ])
        # coerce arg names
        for argnames in wrapped_arg_combos('x', 'y'):
            tests.append(
                    (('SPAM', argnames, '123'),
                     ('SPAM', ('x', 'y'), '123')))
        # missing name, body
        for name in ('', ' ', None, StrProxy(' '), ()):
            for argnames in (None, ()):
                for body in ('', ' ', None, StrProxy(' '), ()):
                    tests.append(
                            ((name, argnames, body),
                             (None, (), None)))
        # missing args
        tests.extend([
            (('SPAM', None, '123'),
             ('SPAM', (), '123')),
            (('SPAM', (), '123'),
             ('SPAM', (), '123')),
            ])
        # missing arg names
        for arg in ('', ' ', None, StrProxy(' '), ()):
            tests.append(
                    (('SPAM', (arg,), '123'),
                     ('SPAM', (None,), '123')))
        tests.extend([
            (('SPAM', ('x', '', 'z'), '123'),
             ('SPAM', ('x', None, 'z'), '123')),
            ])
        # whitespace
        tests.extend([
            ((' SPAM ', (' x ', ' y '), ' 123 '),
             ('SPAM', ('x', 'y'), '123')),
            (('SPAM', 'x, y', '123'),
             ('SPAM', ('x', 'y'), '123')),
            ])

        for args, expected in tests:
            with self.subTest(args):
                d = Macro(*args)

                self.assertEqual(d[1:], expected)
                for i, exp in enumerate(expected, start=1):
                    if i == 2:
                        self.assertIs(type(d[i]), tuple)
                    elif exp is not None:
                        self.assertIs(type(d[i]), str)

    def test_init_bad_args(self):
        tests = [
            ('SPAM', StrProxy('x'), '123'),
            ('SPAM', object(), '123'),
            ]
        for args in tests:
            with self.subTest(args):
                with self.assertRaises(TypeError):
                    Macro(*args)

    def test_valid(self):
        tests = [
            # unusual name
            ('SPAM', ('x', 'y'), 'run(x, y)'),
            ('_SPAM_', ('x', 'y'), 'run(x, y)'),
            ('X_1', ('x', 'y'), 'run(x, y)'),
            # unusual args
            ('SPAM', (), 'run(x, y)'),
            ('SPAM', ('_x_', 'y_1'), 'run(x, y)'),
            ('SPAM', 'x', 'run(x, y)'),
            ('SPAM', 'x, y', 'run(x, y)'),
            # unusual body
            ('SPAM', ('x', 'y'), None),
            ]
        for args in tests:
            with self.subTest(args):
                directive = Macro(*args)

                directive.validate()

    def test_invalid(self):
        tests = [
            # invalid name
            ((None, ('x', 'y'), '123'), TypeError),
            (('_', ('x', 'y'), '123'), ValueError),
            (('1', ('x', 'y'), '123'), ValueError),
            (('_1', ('x', 'y'), '123'), ValueError),
            # invalid args
            (('SPAM', (None, 'y'), '123'), ValueError),
            (('SPAM', ('x', '_'), '123'), ValueError),
            (('SPAM', ('x', '1'), '123'), ValueError),
            (('SPAM', ('x', '_1_'), '123'), ValueError),
            # There is no invalid body (including None).
            ]
        for args, exctype in tests:
            with self.subTest(args):
                directive = Macro(*args)

                with self.assertRaises(exctype):
                    directive.validate()


class IfDirectiveTests(unittest.TestCase):

    def test_type(self):
        directive = IfDirective('if', '1')

        self.assertIs(type(directive), IfDirective)
        self.assertIsInstance(directive, PreprocessorDirective)

    def test_attrs(self):
        d = IfDirective('if', '1')
        kind, condition = d.kind, d.condition

        self.assertEqual(kind, 'if')
        self.assertEqual(condition, '1')
        #self.assertEqual(condition, (ArithmeticCondition('1'),))

    def test_text(self):
        tests = [
            (('if', 'defined(SPAM) && 1 || (EGGS > 3 && defined(HAM))'),
             'defined(SPAM) && 1 || (EGGS > 3 && defined(HAM))'),
            ]
        for kind in IfDirective.KINDS:
            tests.append(
                    ((kind, 'SPAM'), 'SPAM'))
        for args, expected in tests:
            with self.subTest(args):
                d = IfDirective(*args)
                text = d.text

                self.assertEqual(text, expected)

    def test_iter(self):
        kind, condition = IfDirective('if', '1')

        self.assertEqual(kind, 'if')
        self.assertEqual(condition, '1')
        #self.assertEqual(condition, (ArithmeticCondition('1'),))

    #def test_complex_conditions(self):
    #    ...

    def test_coerce(self):
        tests = []
        for kind in IfDirective.KINDS:
            if kind == 'ifdef':
                cond = 'defined(SPAM)'
            elif kind == 'ifndef':
                cond = '! defined(SPAM)'
            else:
                cond = 'SPAM'
            for args in wrapped_arg_combos(kind, 'SPAM'):
                tests.append((args, (kind, cond)))
            tests.extend([
                ((' ' + kind + ' ', ' SPAM '), (kind, cond)),
                ])
            for raw in ('', ' ', None, StrProxy(' '), ()):
                tests.append(((kind, raw), (kind, None)))
        for kind in ('', ' ', None, StrProxy(' '), ()):
            tests.append(((kind, 'SPAM'), (None, 'SPAM')))
        for args, expected in tests:
            with self.subTest(args):
                d = IfDirective(*args)

                self.assertEqual(tuple(d), expected)
                for i, exp in enumerate(expected):
                    if exp is not None:
                        self.assertIs(type(d[i]), str)

    def test_valid(self):
        tests = []
        for kind in IfDirective.KINDS:
            tests.extend([
                (kind, 'SPAM'),
                (kind, '_SPAM_'),
                (kind, 'X_1'),
                (kind, '()'),
                (kind, '--'),
                (kind, '???'),
                ])
        for args in tests:
            with self.subTest(args):
                directive = IfDirective(*args)

                directive.validate()

    def test_invalid(self):
        tests = []
        # kind
        tests.extend([
            ((None, 'SPAM'), TypeError),
            (('_', 'SPAM'), ValueError),
            (('-', 'SPAM'), ValueError),
            (('spam', 'SPAM'), ValueError),
            ])
        for kind in PreprocessorDirective.KINDS:
            if kind in IfDirective.KINDS:
                continue
            tests.append(
                ((kind, 'SPAM'), ValueError))
        # condition
        for kind in IfDirective.KINDS:
            tests.extend([
                ((kind, None), TypeError),
                # Any other condition is valid.
                ])
        for args, exctype in tests:
            with self.subTest(args):
                directive = IfDirective(*args)

                with self.assertRaises(exctype):
                    directive.validate()


class IncludeTests(unittest.TestCase):

    def test_type(self):
        directive = Include('<stdio>')

        self.assertIs(type(directive), Include)
        self.assertIsInstance(directive, PreprocessorDirective)

    def test_attrs(self):
        d = Include('<stdio>')
        kind, file, text = d.kind, d.file, d.text

        self.assertEqual(kind, 'include')
        self.assertEqual(file, '<stdio>')
        self.assertEqual(text, '<stdio>')

    def test_iter(self):
        kind, file = Include('<stdio>')

        self.assertEqual(kind, 'include')
        self.assertEqual(file, '<stdio>')

    def test_coerce(self):
        tests = []
        for arg, in wrapped_arg_combos('<stdio>'):
            tests.append((arg, '<stdio>'))
        tests.extend([
            (' <stdio> ', '<stdio>'),
            ])
        for arg in ('', ' ', None, StrProxy(' '), ()):
            tests.append((arg, None ))
        for arg, expected in tests:
            with self.subTest(arg):
                _, file = Include(arg)

                self.assertEqual(file, expected)
                if expected is not None:
                    self.assertIs(type(file), str)

    def test_valid(self):
        tests = [
            '<stdio>',
            '"spam.h"',
            '"internal/pycore_pystate.h"',
            ]
        for arg in tests:
            with self.subTest(arg):
                directive = Include(arg)

                directive.validate()

    def test_invalid(self):
        tests = [
            (None, TypeError),
            # We currently don't check the file.
            ]
        for arg, exctype in tests:
            with self.subTest(arg):
                directive = Include(arg)

                with self.assertRaises(exctype):
                    directive.validate()


class OtherDirectiveTests(unittest.TestCase):

    def test_type(self):
        directive = OtherDirective('undef', 'SPAM')

        self.assertIs(type(directive), OtherDirective)
        self.assertIsInstance(directive, PreprocessorDirective)

    def test_attrs(self):
        d = OtherDirective('undef', 'SPAM')
        kind, text = d.kind, d.text

        self.assertEqual(kind, 'undef')
        self.assertEqual(text, 'SPAM')

    def test_iter(self):
        kind, text = OtherDirective('undef', 'SPAM')

        self.assertEqual(kind, 'undef')
        self.assertEqual(text, 'SPAM')

    def test_coerce(self):
        tests = []
        for kind in OtherDirective.KINDS:
            if kind in ('else', 'endif'):
                continue
            for args in wrapped_arg_combos(kind, '...'):
                tests.append((args, (kind, '...')))
            tests.extend([
                ((' ' + kind + ' ', ' ... '), (kind, '...')),
                ])
            for raw in ('', ' ', None, StrProxy(' '), ()):
                tests.append(((kind, raw), (kind, None)))
        for kind in ('else', 'endif'):
            for args in wrapped_arg_combos(kind, None):
                tests.append((args, (kind, None)))
            tests.extend([
                ((' ' + kind + ' ', None), (kind, None)),
                ])
        for kind in ('', ' ', None, StrProxy(' '), ()):
            tests.append(((kind, '...'), (None, '...')))
        for args, expected in tests:
            with self.subTest(args):
                d = OtherDirective(*args)

                self.assertEqual(tuple(d), expected)
                for i, exp in enumerate(expected):
                    if exp is not None:
                        self.assertIs(type(d[i]), str)

    def test_valid(self):
        tests = []
        for kind in OtherDirective.KINDS:
            if kind in ('else', 'endif'):
                continue
            tests.extend([
                (kind, '...'),
                (kind, '???'),
                (kind, 'SPAM'),
                (kind, '1 + 1'),
                ])
        for kind in ('else', 'endif'):
            tests.append((kind, None))
        for args in tests:
            with self.subTest(args):
                directive = OtherDirective(*args)

                directive.validate()

    def test_invalid(self):
        tests = []
        # kind
        tests.extend([
            ((None, '...'), TypeError),
            (('_', '...'), ValueError),
            (('-', '...'), ValueError),
            (('spam', '...'), ValueError),
            ])
        for kind in PreprocessorDirective.KINDS:
            if kind in OtherDirective.KINDS:
                continue
            tests.append(
                ((kind, None), ValueError))
        # text
        for kind in OtherDirective.KINDS:
            if kind in ('else', 'endif'):
                tests.extend([
                    # Any text is invalid.
                    ((kind, 'SPAM'), ValueError),
                    ((kind, '...'), ValueError),
                    ])
            else:
                tests.extend([
                    ((kind, None), TypeError),
                    # Any other text is valid.
                    ])
        for args, exctype in tests:
            with self.subTest(args):
                directive = OtherDirective(*args)

                with self.assertRaises(exctype):
                    directive.validate()



"""Parser for the Python metagrammar"""

import io
import tokenize  # from stdlib

from .automata import NFA, NFAState


class GrammarParser:
    """Parser for Python grammar files."""

    _translation_table = {
        tokenize.NAME: "NAME",
        tokenize.STRING: "STRING",
        tokenize.NEWLINE: "NEWLINE",
        tokenize.NL: "NL",
        tokenize.OP: "OP",
        tokenize.ENDMARKER: "ENDMARKER",
        tokenize.COMMENT: "COMMENT",
    }

    def __init__(self, grammar):
        self.grammar = grammar
        grammar_adaptor = io.StringIO(grammar)
        self.generator = tokenize.generate_tokens(grammar_adaptor.readline)
        self._gettoken()  # Initialize lookahead
        self._current_rule_name = None

    def parse(self):
        """Turn the grammar into a collection of NFAs"""
        # grammar: (NEWLINE | rule)* ENDMARKER
        while self.type != tokenize.ENDMARKER:
            while self.type == tokenize.NEWLINE:
                self._gettoken()
            # rule: NAME ':' rhs NEWLINE
            self._current_rule_name = self._expect(tokenize.NAME)
            self._expect(tokenize.OP, ":")
            a, z = self._parse_rhs()
            self._expect(tokenize.NEWLINE)

            yield NFA(a, z)

    def _parse_rhs(self):
        # rhs: items ('|' items)*
        a, z = self._parse_items()
        if self.value != "|":
            return a, z
        else:
            aa = NFAState(self._current_rule_name)
            zz = NFAState(self._current_rule_name)
            while True:
                # Allow to transit directly to the previous state and connect the end of the
                # previous state to the end of the current one, effectively allowing to skip
                # the current state.
                aa.add_arc(a)
                z.add_arc(zz)
                if self.value != "|":
                    break

                self._gettoken()
                a, z = self._parse_items()
            return aa, zz

    def _parse_items(self):
        # items: item+
        a, b = self._parse_item()
        while self.type in (tokenize.NAME, tokenize.STRING) or self.value in ("(", "["):
            c, d = self._parse_item()
            # Allow a transition between the end of the previous state
            # and the beginning of the new one, connecting all the items
            # together. In this way we can only reach the end if we visit
            # all the items.
            b.add_arc(c)
            b = d
        return a, b

    def _parse_item(self):
        # item: '[' rhs ']' | atom ['+' | '*']
        if self.value == "[":
            self._gettoken()
            a, z = self._parse_rhs()
            self._expect(tokenize.OP, "]")
            # Make a transition from the beginning to the end so it is possible to
            # advance for free to the next state of this item # without consuming
            # anything from the rhs.
            a.add_arc(z)
            return a, z
        else:
            a, z = self._parse_atom()
            value = self.value
            if value not in ("+", "*"):
                return a, z
            self._gettoken()
            z.add_arc(a)
            if value == "+":
                # Create a cycle to the beginning so we go back to the old state in this
                # item and repeat.
                return a, z
            else:
                # The end state is the same as the beginning, so we can cycle arbitrarily
                # and end in the beginning if necessary.
                return a, a

    def _parse_atom(self):
        # atom: '(' rhs ')' | NAME | STRING
        if self.value == "(":
            self._gettoken()
            a, z = self._parse_rhs()
            self._expect(tokenize.OP, ")")
            return a, z
        elif self.type in (tokenize.NAME, tokenize.STRING):
            a = NFAState(self._current_rule_name)
            z = NFAState(self._current_rule_name)
            # We can transit to the next state only if we consume the value.
            a.add_arc(z, self.value)
            self._gettoken()
            return a, z
        else:
            self._raise_error(
                "expected (...) or NAME or STRING, got {} ({})",
                self._translation_table.get(self.type, self.type),
                self.value,
            )

    def _expect(self, type_, value=None):
        if self.type != type_:
            self._raise_error(
                "expected {}, got {} ({})",
                self._translation_table.get(type_, type_),
                self._translation_table.get(self.type, self.type),
                self.value,
            )
        if value is not None and self.value != value:
            self._raise_error("expected {}, got {}", value, self.value)
        value = self.value
        self._gettoken()
        return value

    def _gettoken(self):
        tup = next(self.generator)
        while tup[0] in (tokenize.COMMENT, tokenize.NL):
            tup = next(self.generator)
        self.type, self.value, self.begin, self.end, self.line = tup

    def _raise_error(self, msg, *args):
        if args:
            try:
                msg = msg.format(*args)
            except Exception:
                msg = " ".join([msg] + list(map(str, args)))
        line = self.grammar.splitlines()[self.begin[0] - 1]
        raise SyntaxError(msg, ("<grammar>", self.begin[0], self.begin[1], line))



from collections import namedtuple

from c_analyzer.common.info import ID
from c_analyzer.common.util import classonly, _NTBase


class Symbol(_NTBase, namedtuple('Symbol', 'id kind external')):
    """Info for a single compilation symbol."""

    __slots__ = ()

    class KIND:
        VARIABLE = 'variable'
        FUNCTION = 'function'
        OTHER = 'other'

    @classonly
    def from_name(cls, name, filename=None, kind=KIND.VARIABLE, external=None):
        """Return a new symbol based on the given name."""
        id = ID(filename, None, name)
        return cls(id, kind, external)

    def __new__(cls, id, kind=KIND.VARIABLE, external=None):
        self = super().__new__(
                cls,
                id=ID.from_raw(id),
                kind=str(kind) if kind else None,
                external=bool(external) if external is not None else None,
                )
        return self

    def __hash__(self):
        return hash(self.id)

    def __getattr__(self, name):
        return getattr(self.id, name)

    def validate(self):
        """Fail if the object is invalid (i.e. init with bad data)."""
        if not self.id:
            raise TypeError('missing id')
        else:
            self.id.validate()

        if not self.kind:
            raise TypeError('missing kind')
        elif self.kind not in vars(self.KIND).values():
            raise ValueError(f'unsupported kind {self.kind}')

        if self.external is None:
            raise TypeError('missing external')



import os.path
import re

from c_analyzer.common.info import ID
from c_analyzer.common.util import read_tsv, write_tsv

from . import DATA_DIR

# XXX need tests:
# * generate / script


IGNORED_FILE = os.path.join(DATA_DIR, 'ignored.tsv')

IGNORED_COLUMNS = ('filename', 'funcname', 'name', 'kind', 'reason')
IGNORED_HEADER = '\t'.join(IGNORED_COLUMNS)

# XXX Move these to ignored.tsv.
IGNORED = {
        # global
        'PyImport_FrozenModules': 'process-global',
        'M___hello__': 'process-global',
        'inittab_copy': 'process-global',
        'PyHash_Func': 'process-global',
        '_Py_HashSecret_Initialized': 'process-global',
        '_TARGET_LOCALES': 'process-global',

        # startup (only changed before/during)
        '_PyRuntime': 'runtime startup',
        'runtime_initialized': 'runtime startup',
        'static_arg_parsers': 'runtime startup',
        'orig_argv': 'runtime startup',
        'opt_ptr': 'runtime startup',
        '_preinit_warnoptions': 'runtime startup',
        '_Py_StandardStreamEncoding': 'runtime startup',
        'Py_FileSystemDefaultEncoding': 'runtime startup',
        '_Py_StandardStreamErrors': 'runtime startup',
        'Py_FileSystemDefaultEncodeErrors': 'runtime startup',
        'Py_BytesWarningFlag': 'runtime startup',
        'Py_DebugFlag': 'runtime startup',
        'Py_DontWriteBytecodeFlag': 'runtime startup',
        'Py_FrozenFlag': 'runtime startup',
        'Py_HashRandomizationFlag': 'runtime startup',
        'Py_IgnoreEnvironmentFlag': 'runtime startup',
        'Py_InspectFlag': 'runtime startup',
        'Py_InteractiveFlag': 'runtime startup',
        'Py_IsolatedFlag': 'runtime startup',
        'Py_NoSiteFlag': 'runtime startup',
        'Py_NoUserSiteDirectory': 'runtime startup',
        'Py_OptimizeFlag': 'runtime startup',
        'Py_QuietFlag': 'runtime startup',
        'Py_UTF8Mode': 'runtime startup',
        'Py_UnbufferedStdioFlag': 'runtime startup',
        'Py_VerboseFlag': 'runtime startup',
        '_Py_path_config': 'runtime startup',
        '_PyOS_optarg': 'runtime startup',
        '_PyOS_opterr': 'runtime startup',
        '_PyOS_optind': 'runtime startup',
        '_Py_HashSecret': 'runtime startup',

        # REPL
        '_PyOS_ReadlineLock': 'repl',
        '_PyOS_ReadlineTState': 'repl',

        # effectively const
        'tracemalloc_empty_traceback': 'const',
        '_empty_bitmap_node': 'const',
        'posix_constants_pathconf': 'const',
        'posix_constants_confstr': 'const',
        'posix_constants_sysconf': 'const',
        '_PySys_ImplCacheTag': 'const',
        '_PySys_ImplName': 'const',
        'PyImport_Inittab': 'const',
        '_PyImport_DynLoadFiletab': 'const',
        '_PyParser_Grammar': 'const',
        'Py_hexdigits': 'const',
        '_PyImport_Inittab': 'const',
        '_PyByteArray_empty_string': 'const',
        '_PyLong_DigitValue': 'const',
        '_Py_SwappedOp': 'const',
        'PyStructSequence_UnnamedField': 'const',

        # signals are main-thread only
        'faulthandler_handlers': 'signals are main-thread only',
        'user_signals': 'signals are main-thread only',
        'wakeup': 'signals are main-thread only',

        # hacks
        '_PySet_Dummy': 'only used as a placeholder',
        }

BENIGN = 'races here are benign and unlikely'


def is_supported(variable, ignored=None, known=None, *,
                 _ignored=(lambda *a, **k: _is_ignored(*a, **k)),
                 _vartype_okay=(lambda *a, **k: _is_vartype_okay(*a, **k)),
                 ):
    """Return True if the given global variable is okay in CPython."""
    if _ignored(variable,
                ignored and ignored.get('variables')):
        return True
    elif _vartype_okay(variable.vartype,
                       ignored.get('types')):
        return True
    else:
        return False


def _is_ignored(variable, ignoredvars=None, *,
                _IGNORED=IGNORED,
                ):
    """Return the reason if the variable is a supported global.

    Return None if the variable is not a supported global.
    """
    if ignoredvars and (reason := ignoredvars.get(variable.id)):
        return reason

    if variable.funcname is None:
        if reason := _IGNORED.get(variable.name):
            return reason

    # compiler
    if variable.filename == 'Python/graminit.c':
        if variable.vartype.startswith('static state '):
            return 'compiler'
    if variable.filename == 'Python/symtable.c':
        if variable.vartype.startswith('static identifier '):
            return 'compiler'
    if variable.filename == 'Python/Python-ast.c':
        # These should be const.
        if variable.name.endswith('_field'):
            return 'compiler'
        if variable.name.endswith('_attribute'):
            return 'compiler'

    # other
    if variable.filename == 'Python/dtoa.c':
        # guarded by lock?
        if variable.name in ('p5s', 'freelist'):
            return 'dtoa is thread-safe?'
        if variable.name in ('private_mem', 'pmem_next'):
            return 'dtoa is thread-safe?'
    if variable.filename == 'Python/thread.c':
        # Threads do not become an issue until after these have been set
        # and these never get changed after that.
        if variable.name in ('initialized', 'thread_debug'):
            return 'thread-safe'
    if variable.filename == 'Python/getversion.c':
        if variable.name == 'version':
            # Races are benign here, as well as unlikely.
            return BENIGN
    if variable.filename == 'Python/fileutils.c':
        if variable.name == 'force_ascii':
            return BENIGN
        if variable.name == 'ioctl_works':
            return BENIGN
        if variable.name == '_Py_open_cloexec_works':
            return BENIGN
    if variable.filename == 'Python/codecs.c':
        if variable.name == 'ucnhash_CAPI':
            return BENIGN
    if variable.filename == 'Python/bootstrap_hash.c':
        if variable.name == 'getrandom_works':
            return BENIGN
    if variable.filename == 'Objects/unicodeobject.c':
        if variable.name == 'ucnhash_CAPI':
            return BENIGN
        if variable.name == 'bloom_linebreak':
            # *mostly* benign
            return BENIGN
    if variable.filename == 'Modules/getbuildinfo.c':
        if variable.name == 'buildinfo':
            # The static is used for pre-allocation.
            return BENIGN
    if variable.filename == 'Modules/posixmodule.c':
        if variable.name == 'ticks_per_second':
            return BENIGN
        if variable.name == 'dup3_works':
            return BENIGN
    if variable.filename == 'Modules/timemodule.c':
        if variable.name == 'ticks_per_second':
            return BENIGN
    if variable.filename == 'Objects/longobject.c':
        if variable.name == 'log_base_BASE':
            return BENIGN
        if variable.name == 'convwidth_base':
            return BENIGN
        if variable.name == 'convmultmax_base':
            return BENIGN

    return None


def _is_vartype_okay(vartype, ignoredtypes=None):
    if _is_object(vartype):
        return None

    if vartype.startswith('static const '):
        return 'const'
    if vartype.startswith('const '):
        return 'const'

    # components for TypeObject definitions
    for name in ('PyMethodDef', 'PyGetSetDef', 'PyMemberDef'):
        if name in vartype:
            return 'const'
    for name in ('PyNumberMethods', 'PySequenceMethods', 'PyMappingMethods',
                 'PyBufferProcs', 'PyAsyncMethods'):
        if name in vartype:
            return 'const'
    for name in ('slotdef', 'newfunc'):
        if name in vartype:
            return 'const'

    # structseq
    for name in ('PyStructSequence_Desc', 'PyStructSequence_Field'):
        if name in vartype:
            return 'const'

    # other definiitions
    if 'PyModuleDef' in vartype:
        return 'const'

    # thread-safe
    if '_Py_atomic_int' in vartype:
        return 'thread-safe'
    if 'pthread_condattr_t' in vartype:
        return 'thread-safe'

    # startup
    if '_Py_PreInitEntry' in vartype:
        return 'startup'

    # global
#    if 'PyMemAllocatorEx' in vartype:
#        return True

    # others
#    if 'PyThread_type_lock' in vartype:
#        return True

    # XXX ???
    # _Py_tss_t
    # _Py_hashtable_t
    # stack_t
    # _PyUnicode_Name_CAPI

    # functions
    if '(' in vartype and '[' not in vartype:
        return 'function pointer'

    # XXX finish!
    # * allow const values?
    #raise NotImplementedError
    return None


PYOBJECT_RE = re.compile(r'''
        ^
        (
            # must start with "static "
            static \s+
            (
                identifier
            )
            \b
        ) |
        (
            # may start with "static "
            ( static \s+ )?
            (
                .*
                (
                    PyObject |
                    PyTypeObject |
                    _? Py \w+ Object |
                    _PyArg_Parser |
                    _Py_Identifier |
                    traceback_t |
                    PyAsyncGenASend |
                    _PyAsyncGenWrappedValue |
                    PyContext |
                    method_cache_entry
                )
                \b
            ) |
            (
                (
                    _Py_IDENTIFIER |
                    _Py_static_string
                )
                [(]
            )
        )
        ''', re.VERBOSE)


def _is_object(vartype):
    if 'PyDictKeysObject' in vartype:
        return False
    if PYOBJECT_RE.match(vartype):
        return True
    if vartype.endswith((' _Py_FalseStruct', ' _Py_TrueStruct')):
        return True

    # XXX Add more?

    #for part in vartype.split():
    #    # XXX const is automatic True?
    #    if part == 'PyObject' or part.startswith('PyObject['):
    #        return True
    return False


def ignored_from_file(infile, *,
                      _read_tsv=read_tsv,
                      ):
    """Yield a Variable for each ignored var in the file."""
    ignored = {
        'variables': {},
        #'types': {},
        #'constants': {},
        #'macros': {},
        }
    for row in _read_tsv(infile, IGNORED_HEADER):
        filename, funcname, name, kind, reason = row
        if not funcname or funcname == '-':
            funcname = None
        id = ID(filename, funcname, name)
        if kind == 'variable':
            values = ignored['variables']
        else:
            raise ValueError(f'unsupported kind in row {row}')
        values[id] = reason
    return ignored


##################################
# generate

def _get_row(varid, reason):
    return (
            varid.filename,
            varid.funcname or '-',
            varid.name,
            'variable',
            str(reason),
            )


def _get_rows(variables, ignored=None, *,
              _as_row=_get_row,
              _is_ignored=_is_ignored,
              _vartype_okay=_is_vartype_okay,
              ):
    count = 0
    for variable in variables:
        reason = _is_ignored(variable,
                             ignored and ignored.get('variables'),
                             )
        if not reason:
            reason = _vartype_okay(variable.vartype,
                                   ignored and ignored.get('types'))
        if not reason:
            continue

        print(' ', variable, repr(reason))
        yield _as_row(variable.id, reason)
        count += 1
    print(f'total: {count}')


def _generate_ignored_file(variables, filename=None, *,
                           _generate_rows=_get_rows,
                           _write_tsv=write_tsv,
                           ):
    if not filename:
        filename = IGNORED_FILE + '.new'
    rows = _generate_rows(variables)
    _write_tsv(filename, IGNORED_HEADER, rows)


if __name__ == '__main__':
    from cpython import SOURCE_DIRS
    from cpython.known import (
        from_file as known_from_file,
        DATA_FILE as KNOWN_FILE,
        )
    # XXX This is wrong!
    from . import find
    known = known_from_file(KNOWN_FILE)
    knownvars = (known or {}).get('variables')
    variables = find.globals_from_binary(knownvars=knownvars,
                                         dirnames=SOURCE_DIRS)

    _generate_ignored_file(variables)



import sqlite3
import hashlib

def md5sum(t):
    return hashlib.md5(t).hexdigest()

con = sqlite3.connect(":memory:")
con.create_function("md5", 1, md5sum)
cur = con.cursor()
cur.execute("select md5(?)", (b"foo",))
print(cur.fetchone()[0])

con.close()



import string
import unittest

from ..util import PseudoStr, StrProxy, Object
from .. import tool_imports_for_tests
with tool_imports_for_tests():
    from c_analyzer.common.info import ID
    from c_analyzer.symbols.info import Symbol


class SymbolTests(unittest.TestCase):

    VALID_ARGS = (
            ID('x/y/z/spam.c', 'func', 'eggs'),
            Symbol.KIND.VARIABLE,
            False,
            )
    VALID_KWARGS = dict(zip(Symbol._fields, VALID_ARGS))
    VALID_EXPECTED = VALID_ARGS

    def test_init_typical_binary_local(self):
        id = ID(None, None, 'spam')
        symbol = Symbol(
                id=id,
                kind=Symbol.KIND.VARIABLE,
                external=False,
                )

        self.assertEqual(symbol, (
            id,
            Symbol.KIND.VARIABLE,
            False,
            ))

    def test_init_typical_binary_global(self):
        id = ID('Python/ceval.c', None, 'spam')
        symbol = Symbol(
                id=id,
                kind=Symbol.KIND.VARIABLE,
                external=False,
                )

        self.assertEqual(symbol, (
            id,
            Symbol.KIND.VARIABLE,
            False,
            ))

    def test_init_coercion(self):
        tests = [
            ('str subclass',
             dict(
                 id=PseudoStr('eggs'),
                 kind=PseudoStr('variable'),
                 external=0,
                 ),
             (ID(None, None, 'eggs'),
              Symbol.KIND.VARIABLE,
              False,
              )),
            ('with filename',
             dict(
                 id=('x/y/z/spam.c', 'eggs'),
                 kind=PseudoStr('variable'),
                 external=0,
                 ),
             (ID('x/y/z/spam.c', None, 'eggs'),
              Symbol.KIND.VARIABLE,
              False,
              )),
            ('non-str 1',
             dict(
                 id=('a', 'b', 'c'),
                 kind=StrProxy('variable'),
                 external=0,
                 ),
             (ID('a', 'b', 'c'),
              Symbol.KIND.VARIABLE,
              False,
              )),
            ('non-str 2',
             dict(
                 id=('a', 'b', 'c'),
                 kind=Object(),
                 external=0,
                 ),
             (ID('a', 'b', 'c'),
              '<object>',
              False,
              )),
            ]
        for summary, kwargs, expected in tests:
            with self.subTest(summary):
                symbol = Symbol(**kwargs)

                for field in Symbol._fields:
                    value = getattr(symbol, field)
                    if field == 'external':
                        self.assertIs(type(value), bool)
                    elif field == 'id':
                        self.assertIs(type(value), ID)
                    else:
                        self.assertIs(type(value), str)
                self.assertEqual(tuple(symbol), expected)

    def test_init_all_missing(self):
        id = ID(None, None, 'spam')

        symbol = Symbol(id)

        self.assertEqual(symbol, (
            id,
            Symbol.KIND.VARIABLE,
            None,
            ))

    def test_fields(self):
        id = ID('z', 'x', 'a')

        symbol = Symbol(id, 'b', False)

        self.assertEqual(symbol.id, id)
        self.assertEqual(symbol.kind, 'b')
        self.assertIs(symbol.external, False)

    def test___getattr__(self):
        id = ID('z', 'x', 'a')
        symbol = Symbol(id, 'b', False)

        filename = symbol.filename
        funcname = symbol.funcname
        name = symbol.name

        self.assertEqual(filename, 'z')
        self.assertEqual(funcname, 'x')
        self.assertEqual(name, 'a')

    def test_validate_typical(self):
        id = ID('z', 'x', 'a')

        symbol = Symbol(
                id=id,
                kind=Symbol.KIND.VARIABLE,
                external=False,
                )

        symbol.validate()  # This does not fail.

    def test_validate_missing_field(self):
        for field in Symbol._fields:
            with self.subTest(field):
                symbol = Symbol(**self.VALID_KWARGS)
                symbol = symbol._replace(**{field: None})

                with self.assertRaises(TypeError):
                    symbol.validate()

    def test_validate_bad_field(self):
        badch = tuple(c for c in string.punctuation + string.digits)
        notnames = (
                '1a',
                'a.b',
                'a-b',
                '&a',
                'a++',
                ) + badch
        tests = [
            ('id', notnames),
            ('kind', ('bogus',)),
            ]
        seen = set()
        for field, invalid in tests:
            for value in invalid:
                if field != 'kind':
                    seen.add(value)
                with self.subTest(f'{field}={value!r}'):
                    symbol = Symbol(**self.VALID_KWARGS)
                    symbol = symbol._replace(**{field: value})

                    with self.assertRaises(ValueError):
                        symbol.validate()

        for field, invalid in tests:
            if field == 'kind':
                continue
            valid = seen - set(invalid)
            for value in valid:
                with self.subTest(f'{field}={value!r}'):
                    symbol = Symbol(**self.VALID_KWARGS)
                    symbol = symbol._replace(**{field: value})

                    symbol.validate()  # This does not fail.



import ast
import _peg_parser as peg_parser
import unittest
from typing import Any, Union, Iterable, Tuple
from textwrap import dedent
from test import support


TEST_CASES = [
    ('annotated_assignment', 'x: int = 42'),
    ('annotated_assignment_with_tuple', 'x: tuple = 1, 2'),
    ('annotated_assignment_with_parens', '(paren): int = 3+2'),
    ('annotated_assignment_with_yield', 'x: int = yield 42'),
    ('annotated_no_assignment', 'x: int'),
    ('annotation_with_multiple_parens', '((parens)): int'),
    ('annotation_with_parens', '(parens): int'),
    ('annotated_assignment_with_attr', 'a.b: int'),
    ('annotated_assignment_with_subscript', 'a[b]: int'),
    ('annotated_assignment_with_attr_and_parens', '(a.b): int'),
    ('annotated_assignment_with_subscript_and_parens', '(a[b]): int'),
    ('assert', 'assert a'),
    ('assert_message', 'assert a, b'),
    ('assignment_false', 'a = False'),
    ('assignment_none', 'a = None'),
    ('assignment_true', 'a = True'),
    ('assignment_paren', '(a) = 42'),
    ('assignment_paren_multiple', '(a, b) = (0, 1)'),
    ('asyncfor',
     '''
        async for i in a:
            pass
     '''),
    ('attribute_call', 'a.b()'),
    ('attribute_multiple_names', 'abcd.efg.hij'),
    ('attribute_simple', 'a.b'),
    ('attributes_subscript', 'a.b[0]'),
    ('augmented_assignment', 'x += 42'),
    ('augmented_assignment_attribute', 'a.b.c += 42'),
    ('augmented_assignment_paren', '(x) += 42'),
    ('augmented_assignment_paren_subscript', '(x[0]) -= 42'),
    ('binop_add', '1 + 1'),
    ('binop_add_multiple', '1 + 1 + 1 + 1'),
    ('binop_all', '1 + 2 * 5 + 3 ** 2 - -3'),
    ('binop_boolop_comp', '1 + 1 == 2 or 1 + 1 == 3 and not b'),
    ('boolop_or', 'a or b'),
    ('boolop_or_multiple', 'a or b or c'),
    ('class_def_bases',
     '''
        class C(A, B):
            pass
     '''),
    ('class_def_decorators',
     '''
        @a
        class C:
            pass
     '''),
    ('class_def_decorator_with_expression',
     '''
        @lambda x: 42
        class C:
            pass
     '''),
    ('class_def_decorator_with_expression_and_walrus',
     '''
        @x:=lambda x: 42
        class C:
            pass
     '''),

    ('class_def_keywords',
     '''
        class C(keyword=a+b, **c):
            pass
     '''),
    ('class_def_mixed',
     '''
        class C(A, B, keyword=0, **a):
            pass
     '''),
    ('class_def_simple',
     '''
        class C:
            pass
     '''),
    ('class_def_starred_and_kwarg',
     '''
        class C(A, B, *x, **y):
            pass
     '''),
    ('class_def_starred_in_kwargs',
     '''
        class C(A, x=2, *[B, C], y=3):
            pass
     '''),
    ('call_attribute', 'f().b'),
    ('call_genexp', 'f(i for i in a)'),
    ('call_mixed_args', 'f(a, b, *c, **d)'),
    ('call_mixed_args_named', 'f(a, b, *c, d=4, **v)'),
    ('call_one_arg', 'f(a)'),
    ('call_posarg_genexp', 'f(a, (i for i in a))'),
    ('call_simple', 'f()'),
    ('call_subscript', 'f()[0]'),
    ('comp', 'a == b'),
    ('comp_multiple', 'a == b == c'),
    ('comp_paren_end', 'a == (b-1)'),
    ('comp_paren_start', '(a-1) == b'),
    ('decorator',
     '''
        @a
        def f():
            pass
     '''),
    ('decorator_async',
     '''
        @a
        async def d():
            pass
     '''),
    ('decorator_with_expression',
     '''
        @lambda x: 42
        def f():
            pass
     '''),
    ('decorator_with_expression_and_walrus',
     '''
        @x:=lambda x: 42
        def f():
            pass
     '''),
    ('del_attribute', 'del a.b'),
    ('del_call_attribute', 'del a().c'),
    ('del_call_genexp_attribute', 'del a(i for i in b).c'),
    ('del_empty', 'del()'),
    ('del_list', 'del a, [b, c]'),
    ('del_mixed', 'del a[0].b().c'),
    ('del_multiple', 'del a, b'),
    ('del_multiple_calls_attribute', 'del a()().b'),
    ('del_paren', 'del(a,b)'),
    ('del_paren_single_target', 'del(a)'),
    ('del_subscript_attribute', 'del a[0].b'),
    ('del_tuple', 'del a, (b, c)'),
    ('delete', 'del a'),
    ('dict',
     '''
        {
            a: 1,
            b: 2,
            c: 3
        }
     '''),
    ('dict_comp', '{x:1 for x in a}'),
    ('dict_comp_if', '{x:1+2 for x in a if b}'),
    ('dict_empty', '{}'),
    ('empty_line_after_linecont',
     r'''
        pass
        \

        pass
     '''),
    ('for',
     '''
        for i in a:
            pass
     '''),
    ('for_else',
     '''
        for i in a:
            pass
        else:
            pass
     '''),
    ('for_star_target_in_paren', 'for (a) in b: pass'),
    ('for_star_targets_attribute', 'for a.b in c: pass'),
    ('for_star_targets_call_attribute', 'for a().c in b: pass'),
    ('for_star_targets_empty', 'for () in a: pass'),
    ('for_star_targets_mixed', 'for a[0].b().c in d: pass'),
    ('for_star_targets_mixed_starred',
     '''
        for a, *b, (c, d) in e:
            pass
     '''),
    ('for_star_targets_multiple', 'for a, b in c: pass'),
    ('for_star_targets_nested_starred', 'for *[*a] in b: pass'),
    ('for_star_targets_starred', 'for *a in b: pass'),
    ('for_star_targets_subscript_attribute', 'for a[0].b in c: pass'),
    ('for_star_targets_trailing_comma',
     '''
        for a, (b, c), in d:
            pass
     '''),
    ('for_star_targets_tuple', 'for a, (b, c) in d: pass'),
    ('for_underscore',
     '''
        for _ in a:
            pass
     '''),
    ('function_return_type',
     '''
        def f() -> Any:
            pass
     '''),
    ('f-string_slice', "f'{x[2]}'"),
    ('f-string_slice_upper', "f'{x[2:3]}'"),
    ('f-string_slice_step', "f'{x[2:3:-2]}'"),
    ('f-string_constant', "f'{42}'"),
    ('f-string_boolop', "f'{x and y}'"),
    ('f-string_named_expr', "f'{(x:=42)}'"),
    ('f-string_binop', "f'{x+y}'"),
    ('f-string_unaryop', "f'{not x}'"),
    ('f-string_lambda', "f'{(lambda x, /, y, y2=42 , *z, k1, k2=34, **k3: 42)}'"),
    ('f-string_lambda_call', "f'{(lambda: 2)(2)}'"),
    ('f-string_ifexpr', "f'{x if y else z}'"),
    ('f-string_dict', "f'{ {2:34, 3:34} }'"),
    ('f-string_set', "f'{ {2,-45} }'"),
    ('f-string_list', "f'{ [2,-45] }'"),
    ('f-string_tuple', "f'{ (2,-45) }'"),
    ('f-string_listcomp', "f'{[x for x in y if z]}'"),
    ('f-string_setcomp', "f'{ {x for x in y if z} }'"),
    ('f-string_dictcomp', "f'{ {x:x for x in y if z} }'"),
    ('f-string_genexpr', "f'{ (x for x in y if z) }'"),
    ('f-string_yield', "f'{ (yield x) }'"),
    ('f-string_yieldfrom', "f'{ (yield from x) }'"),
    ('f-string_await', "f'{ await x }'"),
    ('f-string_compare', "f'{ x == y }'"),
    ('f-string_call', "f'{ f(x,y,z) }'"),
    ('f-string_attribute', "f'{ f.x.y.z }'"),
    ('f-string_starred', "f'{ *x, }'"),
    ('f-string_doublestarred', "f'{ {**x} }'"),
    ('f-string_escape_brace', "f'{{Escape'"),
    ('f-string_escape_closing_brace', "f'Escape}}'"),
    ('f-string_repr', "f'{a!r}'"),
    ('f-string_str', "f'{a!s}'"),
    ('f-string_ascii', "f'{a!a}'"),
    ('f-string_debug', "f'{a=}'"),
    ('f-string_padding', "f'{a:03d}'"),
    ('f-string_multiline',
     """
        f'''
        {hello}
        '''
     """),
    ('f-string_multiline_in_expr',
     """
        f'''
        {
        hello
        }
        '''
     """),
    ('f-string_multiline_in_call',
     """
        f'''
        {f(
            a, b, c
        )}
        '''
     """),
    ('global', 'global a, b'),
    ('group', '(yield a)'),
    ('if_elif',
     '''
        if a:
            pass
        elif b:
            pass
     '''),
    ('if_elif_elif',
     '''
        if a:
            pass
        elif b:
            pass
        elif c:
            pass
     '''),
    ('if_elif_else',
     '''
        if a:
            pass
        elif b:
            pass
        else:
           pass
     '''),
    ('if_else',
     '''
        if a:
            pass
        else:
            pass
     '''),
    ('if_simple', 'if a: pass'),
    ('import', 'import a'),
    ('import_alias', 'import a as b'),
    ('import_dotted', 'import a.b'),
    ('import_dotted_alias', 'import a.b as c'),
    ('import_dotted_multichar', 'import ab.cd'),
    ('import_from', 'from a import b'),
    ('import_from_alias', 'from a import b as c'),
    ('import_from_dotted', 'from a.b import c'),
    ('import_from_dotted_alias', 'from a.b import c as d'),
    ('import_from_multiple_aliases', 'from a import b as c, d as e'),
    ('import_from_one_dot', 'from .a import b'),
    ('import_from_one_dot_alias', 'from .a import b as c'),
    ('import_from_star', 'from a import *'),
    ('import_from_three_dots', 'from ...a import b'),
    ('import_from_trailing_comma', 'from a import (b,)'),
    ('kwarg',
     '''
        def f(**a):
            pass
     '''),
    ('kwonly_args',
     '''
        def f(*, a, b):
            pass
     '''),
    ('kwonly_args_with_default',
     '''
        def f(*, a=2, b):
            pass
     '''),
    ('lambda_kwarg', 'lambda **a: 42'),
    ('lambda_kwonly_args', 'lambda *, a, b: 42'),
    ('lambda_kwonly_args_with_default', 'lambda *, a=2, b: 42'),
    ('lambda_mixed_args', 'lambda a, /, b, *, c: 42'),
    ('lambda_mixed_args_with_default', 'lambda a, b=2, /, c=3, *e, f, **g: 42'),
    ('lambda_no_args', 'lambda: 42'),
    ('lambda_pos_args', 'lambda a,b: 42'),
    ('lambda_pos_args_with_default', 'lambda a, b=2: 42'),
    ('lambda_pos_only_args', 'lambda a, /: 42'),
    ('lambda_pos_only_args_with_default', 'lambda a=0, /: 42'),
    ('lambda_pos_posonly_args', 'lambda a, b, /, c, d: 42'),
    ('lambda_pos_posonly_args_with_default', 'lambda a, b=0, /, c=2: 42'),
    ('lambda_vararg', 'lambda *a: 42'),
    ('lambda_vararg_kwonly_args', 'lambda *a, b: 42'),
    ('list', '[1, 2, a]'),
    ('list_comp', '[i for i in a]'),
    ('list_comp_if', '[i for i in a if b]'),
    ('list_trailing_comma', '[1+2, a, 3+4,]'),
    ('mixed_args',
     '''
        def f(a, /, b, *, c):
            pass
     '''),
    ('mixed_args_with_default',
     '''
        def f(a, b=2, /, c=3, *e, f, **g):
            pass
     '''),
    ('multipart_string_bytes', 'b"Hola" b"Hello" b"Bye"'),
    ('multipart_string_triple', '"""Something here""" "and now"'),
    ('multipart_string_different_prefixes', 'u"Something" "Other thing" r"last thing"'),
    ('multiple_assignments', 'x = y = z = 42'),
    ('multiple_assignments_with_yield', 'x = y = z = yield 42'),
    ('multiple_pass',
     '''
        pass; pass
        pass
     '''),
    ('namedexpr', '(x := [1, 2, 3])'),
    ('namedexpr_false', '(x := False)'),
    ('namedexpr_none', '(x := None)'),
    ('namedexpr_true', '(x := True)'),
    ('nonlocal', 'nonlocal a, b'),
    ('number_complex', '-2.234+1j'),
    ('number_float', '-34.2333'),
    ('number_imaginary_literal', '1.1234j'),
    ('number_integer', '-234'),
    ('number_underscores', '1_234_567'),
    ('pass', 'pass'),
    ('pos_args',
     '''
        def f(a, b):
            pass
     '''),
    ('pos_args_with_default',
     '''
        def f(a, b=2):
            pass
     '''),
    ('pos_only_args',
     '''
        def f(a, /):
            pass
     '''),
    ('pos_only_args_with_default',
     '''
        def f(a=0, /):
            pass
     '''),
    ('pos_posonly_args',
     '''
        def f(a, b, /, c, d):
            pass
     '''),
    ('pos_posonly_args_with_default',
     '''
        def f(a, b=0, /, c=2):
            pass
     '''),
    ('primary_mixed', 'a.b.c().d[0]'),
    ('raise', 'raise'),
    ('raise_ellipsis', 'raise ...'),
    ('raise_expr', 'raise a'),
    ('raise_from', 'raise a from b'),
    ('return', 'return'),
    ('return_expr', 'return a'),
    ('set', '{1, 2+4, 3+5}'),
    ('set_comp', '{i for i in a}'),
    ('set_trailing_comma', '{1, 2, 3,}'),
    ('simple_assignment', 'x = 42'),
    ('simple_assignment_with_yield', 'x = yield 42'),
    ('string_bytes', 'b"hello"'),
    ('string_concatenation_bytes', 'b"hello" b"world"'),
    ('string_concatenation_simple', '"abcd" "efgh"'),
    ('string_format_simple', 'f"hello"'),
    ('string_format_with_formatted_value', 'f"hello {world}"'),
    ('string_simple', '"hello"'),
    ('string_unicode', 'u"hello"'),
    ('subscript_attribute', 'a[0].b'),
    ('subscript_call', 'a[b]()'),
    ('subscript_multiple_slices', 'a[0:a:2, 1]'),
    ('subscript_simple', 'a[0]'),
    ('subscript_single_element_tuple', 'a[0,]'),
    ('subscript_trailing_comma', 'a[0, 1, 2,]'),
    ('subscript_tuple', 'a[0, 1, 2]'),
    ('subscript_whole_slice', 'a[0+1:b:c]'),
    ('try_except',
     '''
        try:
            pass
        except:
            pass
     '''),
    ('try_except_else',
     '''
        try:
            pass
        except:
            pass
        else:
            pass
     '''),
    ('try_except_else_finally',
     '''
        try:
            pass
        except:
            pass
        else:
            pass
        finally:
            pass
     '''),
    ('try_except_expr',
     '''
        try:
            pass
        except a:
            pass
     '''),
    ('try_except_expr_target',
     '''
        try:
            pass
        except a as b:
            pass
     '''),
    ('try_except_finally',
     '''
        try:
            pass
        except:
            pass
        finally:
            pass
     '''),
    ('try_finally',
     '''
        try:
            pass
        finally:
            pass
     '''),
    ('unpacking_binop', '[*([1, 2, 3] + [3, 4, 5])]'),
    ('unpacking_call', '[*b()]'),
    ('unpacking_compare', '[*(x < y)]'),
    ('unpacking_constant', '[*3]'),
    ('unpacking_dict', '[*{1: 2, 3: 4}]'),
    ('unpacking_dict_comprehension', '[*{x:y for x,y in z}]'),
    ('unpacking_ifexpr', '[*([1, 2, 3] if x else y)]'),
    ('unpacking_list', '[*[1,2,3]]'),
    ('unpacking_list_comprehension', '[*[x for x in y]]'),
    ('unpacking_namedexpr', '[*(x:=[1, 2, 3])]'),
    ('unpacking_set', '[*{1,2,3}]'),
    ('unpacking_set_comprehension', '[*{x for x in y}]'),
    ('unpacking_string', '[*"myvalue"]'),
    ('unpacking_tuple', '[*(1,2,3)]'),
    ('unpacking_unaryop', '[*(not [1, 2, 3])]'),
    ('unpacking_yield', '[*(yield 42)]'),
    ('unpacking_yieldfrom', '[*(yield from x)]'),
    ('tuple', '(1, 2, 3)'),
    ('vararg',
     '''
        def f(*a):
            pass
     '''),
    ('vararg_kwonly_args',
     '''
        def f(*a, b):
            pass
     '''),
    ('while',
     '''
        while a:
            pass
     '''),
    ('while_else',
     '''
        while a:
            pass
        else:
             pass
    '''),
    ('with',
     '''
        with a:
            pass
     '''),
    ('with_as',
     '''
        with a as b:
            pass
     '''),
    ('with_as_paren',
     '''
        with a as (b):
            pass
     '''),
    ('with_as_empty', 'with a as (): pass'),
    ('with_list_recursive',
     '''
        with a as [x, [y, z]]:
            pass
     '''),
    ('with_tuple_recursive',
     '''
        with a as ((x, y), z):
            pass
     '''),
    ('with_tuple_target',
     '''
        with a as (x, y):
            pass
     '''),
    ('with_list_target',
     '''
        with a as [x, y]:
            pass
     '''),
    ('yield', 'yield'),
    ('yield_expr', 'yield a'),
    ('yield_from', 'yield from a'),
]

FAIL_TEST_CASES = [
    ("annotation_multiple_targets", "(a, b): int = 42"),
    ("annotation_nested_tuple", "((a, b)): int"),
    ("annotation_list", "[a]: int"),
    ("annotation_lambda", "lambda: int = 42"),
    ("annotation_tuple", "(a,): int"),
    ("annotation_tuple_without_paren", "a,: int"),
    ("assignment_keyword", "a = if"),
    ("augmented_assignment_list", "[a, b] += 1"),
    ("augmented_assignment_tuple", "a, b += 1"),
    ("augmented_assignment_tuple_paren", "(a, b) += (1, 2)"),
    ("comprehension_lambda", "(a for a in lambda: b)"),
    ("comprehension_else", "(a for a in b if c else d"),
    ("del_call", "del a()"),
    ("del_call_genexp", "del a(i for i in b)"),
    ("del_subscript_call", "del a[b]()"),
    ("del_attribute_call", "del a.b()"),
    ("del_mixed_call", "del a[0].b().c.d()"),
    ("for_star_targets_call", "for a() in b: pass"),
    ("for_star_targets_subscript_call", "for a[b]() in c: pass"),
    ("for_star_targets_attribute_call", "for a.b() in c: pass"),
    ("for_star_targets_mixed_call", "for a[0].b().c.d() in e: pass"),
    ("for_star_targets_in", "for a, in in b: pass"),
    ("f-string_assignment", "f'{x = 42}'"),
    ("f-string_empty", "f'{}'"),
    ("f-string_function_def", "f'{def f(): pass}'"),
    ("f-string_lambda", "f'{lambda x: 42}'"),
    ("f-string_singe_brace", "f'{'"),
    ("f-string_single_closing_brace", "f'}'"),
    ("from_import_invalid", "from import import a"),
    ("from_import_trailing_comma", "from a import b,"),
    ("import_non_ascii_syntax_error", "import ä £"),
    # This test case checks error paths involving tokens with uninitialized
    # values of col_offset and end_col_offset.
    ("invalid indentation",
     """
     def f():
         a
             a
     """),
    ("not_terminated_string", "a = 'example"),
    ("try_except_attribute_target",
     """
     try:
         pass
     except Exception as a.b:
         pass
     """),
    ("try_except_subscript_target",
     """
     try:
         pass
     except Exception as a[0]:
         pass
     """),
]

FAIL_SPECIALIZED_MESSAGE_CASES = [
    ("f(x, y, z=1, **b, *a", "iterable argument unpacking follows keyword argument unpacking"),
    ("f(x, y=1, *z, **a, b", "positional argument follows keyword argument unpacking"),
    ("f(x, y, z=1, a=2, b", "positional argument follows keyword argument"),
    ("True = 1", "cannot assign to True"),
    ("a() = 1", "cannot assign to function call"),
    ("(a, b): int", "only single target (not tuple) can be annotated"),
    ("[a, b]: int", "only single target (not list) can be annotated"),
    ("a(): int", "illegal target for annotation"),
    ("1 += 1", "'literal' is an illegal expression for augmented assignment"),
    ("pass\n    pass", "unexpected indent"),
    ("def f():\npass", "expected an indented block"),
    ("def f(*): pass", "named arguments must follow bare *"),
    ("def f(*,): pass", "named arguments must follow bare *"),
    ("def f(*, **a): pass", "named arguments must follow bare *"),
    ("lambda *: pass", "named arguments must follow bare *"),
    ("lambda *,: pass", "named arguments must follow bare *"),
    ("lambda *, **a: pass", "named arguments must follow bare *"),
    ("f(g()=2", "expression cannot contain assignment, perhaps you meant \"==\"?"),
    ("f(a, b, *c, d.e=2", "expression cannot contain assignment, perhaps you meant \"==\"?"),
    ("f(*a, **b, c=0, d[1]=3)", "expression cannot contain assignment, perhaps you meant \"==\"?"),
]

GOOD_BUT_FAIL_TEST_CASES = [
    ('string_concatenation_format', 'f"{hello} world" f"again {and_again}"'),
    ('string_concatenation_multiple',
     '''
        f"hello" f"{world} again" f"and_again"
     '''),
    ('f-string_multiline_comp',
     """
        f'''
        {(i for i in a
            if b)}
        '''
     """),
]

FSTRINGS_TRACEBACKS = {
    'multiline_fstrings_same_line_with_brace': (
        """
            f'''
            {a$b}
            '''
        """,
        '(a$b)\n',
    ),
    'multiline_fstring_brace_on_next_line': (
        """
            f'''
            {a$b
            }'''
        """,
        '(a$b\n',
    ),
    'multiline_fstring_brace_on_previous_line': (
        """
            f'''
            {
            a$b}'''
        """,
        'a$b)\n',
    ),
}

EXPRESSIONS_TEST_CASES = [
    ("expression_add", "1+1"),
    ("expression_add_2", "a+b"),
    ("expression_call", "f(a, b=2, **kw)"),
    ("expression_tuple", "1, 2, 3"),
    ("expression_tuple_one_value", "1,")
]


def cleanup_source(source: Any) -> str:
    if isinstance(source, str):
        result = dedent(source)
    elif not isinstance(source, (list, tuple)):
        result = "\n".join(source)
    else:
        raise TypeError(f"Invalid type for test source: {source}")
    return result


def prepare_test_cases(
    test_cases: Iterable[Tuple[str, Union[str, Iterable[str]]]]
) -> Tuple[Iterable[str], Iterable[str]]:

    test_ids, _test_sources = zip(*test_cases)
    test_sources = list(_test_sources)
    for index, source in enumerate(test_sources):
        result = cleanup_source(source)
        test_sources[index] = result
    return test_ids, test_sources


TEST_IDS, TEST_SOURCES = prepare_test_cases(TEST_CASES)

GOOD_BUT_FAIL_TEST_IDS, GOOD_BUT_FAIL_SOURCES = prepare_test_cases(
    GOOD_BUT_FAIL_TEST_CASES
)

FAIL_TEST_IDS, FAIL_SOURCES = prepare_test_cases(FAIL_TEST_CASES)

EXPRESSIONS_TEST_IDS, EXPRESSIONS_TEST_SOURCES = prepare_test_cases(
    EXPRESSIONS_TEST_CASES
)


class ASTGenerationTest(unittest.TestCase):
    def test_correct_ast_generation_on_source_files(self) -> None:
        self.maxDiff = None
        for source in TEST_SOURCES:
            actual_ast = peg_parser.parse_string(source)
            expected_ast = peg_parser.parse_string(source, oldparser=True)
            self.assertEqual(
                ast.dump(actual_ast, include_attributes=True),
                ast.dump(expected_ast, include_attributes=True),
                f"Wrong AST generation for source: {source}",
            )

    def test_incorrect_ast_generation_on_source_files(self) -> None:
        for source in FAIL_SOURCES:
            with self.assertRaises(SyntaxError, msg=f"Parsing {source} did not raise an exception"):
                peg_parser.parse_string(source)

    def test_incorrect_ast_generation_with_specialized_errors(self) -> None:
        for source, error_text in FAIL_SPECIALIZED_MESSAGE_CASES:
            exc = IndentationError if "indent" in error_text else SyntaxError
            with self.assertRaises(exc) as se:
                peg_parser.parse_string(source)
            self.assertTrue(
                error_text in se.exception.msg,
                f"Actual error message does not match expexted for {source}"
            )

    @unittest.expectedFailure
    def test_correct_but_known_to_fail_ast_generation_on_source_files(self) -> None:
        for source in GOOD_BUT_FAIL_SOURCES:
            actual_ast = peg_parser.parse_string(source)
            expected_ast = peg_parser.parse_string(source, oldparser=True)
            self.assertEqual(
                ast.dump(actual_ast, include_attributes=True),
                ast.dump(expected_ast, include_attributes=True),
                f"Wrong AST generation for source: {source}",
            )

    def test_correct_ast_generation_without_pos_info(self) -> None:
        for source in GOOD_BUT_FAIL_SOURCES:
            actual_ast = peg_parser.parse_string(source)
            expected_ast = peg_parser.parse_string(source, oldparser=True)
            self.assertEqual(
                ast.dump(actual_ast),
                ast.dump(expected_ast),
                f"Wrong AST generation for source: {source}",
            )

    def test_fstring_parse_error_tracebacks(self) -> None:
        for source, error_text in FSTRINGS_TRACEBACKS.values():
            with self.assertRaises(SyntaxError) as se:
                peg_parser.parse_string(dedent(source))
            self.assertEqual(error_text, se.exception.text)

    def test_correct_ast_generatrion_eval(self) -> None:
        for source in EXPRESSIONS_TEST_SOURCES:
            actual_ast = peg_parser.parse_string(source, mode='eval')
            expected_ast = peg_parser.parse_string(source, mode='eval', oldparser=True)
            self.assertEqual(
                ast.dump(actual_ast, include_attributes=True),
                ast.dump(expected_ast, include_attributes=True),
                f"Wrong AST generation for source: {source}",
            )

    def test_tokenizer_errors_are_propagated(self) -> None:
        n=201
        with self.assertRaisesRegex(SyntaxError, "too many nested parentheses"):
            peg_parser.parse_string(n*'(' + ')'*n)
